from edge import eventlog
from edge.service import react
from .services import stats, event, store, api

from _args import KernelArgs
from _models import Fact


class BaseKernel:
    def __init__(self, api_service, stats_service, react_service, store_service, event_service):
        """
        :type api_service: api.IApiService
        :type stats_service: stats.StatsService
        :type react_service: react.IReactService
        :type store_service: store.IStoreService
        :type event_service: event.IEventService
        """
        self._api_service = api_service
        self._api_service.set_handler(self._handle_query)

        self._react_service = react_service
        self._react_service.set_handler(self._handle_react)
        self._react_counters = stats_service.react

        self._store_service = store_service
        self._event_service = event_service

    def _handle_query(self, query_jnode):
        """
        :type query_jnode: dict
        :rtype: dict
        """
        action = query_jnode.get('action')
        args_jnode = query_jnode.get('args')
        reply_jnode = {'code': 500, 'action': action, 'args': args_jnode}

        if not args_jnode:
            reply_jnode.update({'code': 400, 'text': 'param #/args has invalid value'})
            eventlog.log_w('fail: react-conf <{}>'.format(reply_jnode['text']))

        def exec_react_conf(react_service):
            """
            :type react_service: react.IReactService
            """
            eventlog.log_i('exec')
            try:
                react_service.set_endpoints(args_jnode)
                reply_jnode.update({'code': 200})
                eventlog.log_i('done')
            except react.ReactError:
                eventlog.log_w('fail')

        def exec_store_conf(store_service):
            """
            :type store_service: store.IStoreService
            """
            eventlog.log_d('exec')
            try:
                store_service.setup(args_jnode)
                reply_jnode.update({'code': 200})
                eventlog.log_d('done')
            except store.StoreError:
                eventlog.log_w('fail')

        # dispatch
        if action == 'react-conf':
            exec_react_conf(self._react_service)

        elif action == 'store-conf':
            exec_store_conf(self._store_service)

        else:
            reply_jnode.update({'code': 400, 'text': 'param #/action has invalid value'})
            eventlog.log_w('fail: {}'.format(reply_jnode['text']))

        return reply_jnode

    def _handle_react(self, react_name, react_jnode):
        """
        :type react_name: str
        :type react_jnode: dict
        """
        try:
            self._react_counters.inc_fact_count()
            fact = Fact.unpack(react_name, react_jnode)
        except (KeyError, ValueError) as x:
            eventlog.log_e('fail: unpack {} <{}>'.format(react_name, x.message))
            return

        try:
            fact_ref = self._store_service.save(fact)
            fact_ref and self._event_service.consume_fact(fact_ref, fact)
        except (store.StoreError, event.EventError):
            pass


class Kernel(BaseKernel):
    def __init__(self, twisted_reactor, args):
        """
        :type args: KernelArgs
        """
        stats_service = stats.init_service(
            twisted_reactor, args.stats, bool(args.event))

        event_service = event.init_service(stats_service.event, args.event)
        store_service = store.init_service(stats_service.store, args.store)
        react_service = react.init_service(twisted_reactor, args.react.endpoints)
        api_service = api.init_service(twisted_reactor, args.api)

        def shutdown():
            api.fini_service(api_service)
            react.fini_service(react_service)
            store.fini_service(store_service)
            event.fini_service(event_service)
            stats.fini_service(stats_service)

        twisted_reactor.addSystemEventTrigger('before', 'shutdown', shutdown)

        BaseKernel.__init__(self, api_service, stats_service, react_service, store_service, event_service)
        self._reactor = twisted_reactor
        self._counters = stats_service

    def run(self):
        self._reactor.run()
