import os
import ujson

from twisted.protocols.basic import LineOnlyReceiver
from twisted.protocols.policies import LimitTotalConnectionsFactory
from twisted.internet.protocol import connectionDone
from twisted.internet.error import ConnectionDone

from edge import eventlog
from .types import ApiArgs


class IApiService:
    def __init__(self):
        pass

    def set_handler(self, handler):
        """
        :type handler: (dict) -> dict
        """
        raise NotImplementedError

    def destroy(self):
        pass


class _ApiClientProtocol(LineOnlyReceiver):
    delimiter = b'\0'

    def __init__(self):
        self.timeout_call = None

    def connectionMade(self):
        eventlog.log_d('exec')

    def connectionLost(self, reason=connectionDone):
        if reason.check([ConnectionDone]):
            eventlog.log_e('done: {}'.format(reason))
        else:
            eventlog.log_d('done')

    def lineReceived(self, query_utf8):
        eventlog.log_d('query: {}'.format(query_utf8))

        # noinspection PyBroadException
        try:
            query_jnode = ujson.loads(query_utf8)
            reply_jnode = self.factory.exec_query(query_jnode)
        except Exception as x:
            eventlog.log_x()
            reply_jnode = {
                'code': 500,
                'text': x.message
            }

        reply_utf8 = ujson.dumps(reply_jnode, ensure_ascii=False)
        eventlog.log_d('reply: {}'.format(reply_utf8))
        self.sendLine(reply_utf8)


class _ApiServer(LimitTotalConnectionsFactory, IApiService):
    connectionLimit = 1

    def __init__(self, twisted_reactor, endpoint):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorUNIX
        :type endpoint: str
        """
        IApiService.__init__(self)

        self._handler = None
        self.protocol = _ApiClientProtocol  # override Factory::protocol
        twisted_reactor.listenUNIX(endpoint, self)

    def set_handler(self, handler):
        """
        :type handler: (dict | None) -> dict
        """
        self._handler = handler

    def exec_query(self, query_args):
        """
        :type query_args: dict
        :rtype: dict
        """
        return self._handler(query_args)


def init_service(twisted_reactor, args):
    """
    :type twisted_reactor: twisted.internet.interfaces.IReactorUNIX
    :type args: ApiArgs
    :rtype: IApiService
    """
    if os.path.exists(args.endpoint):
        os.unlink(args.endpoint)

    return _ApiServer(twisted_reactor, args.endpoint)


def fini_service(service):
    """
    :type service: IApiService | _ApiServer
    """
    service.destroy()
