import base64
import datetime
import iso8601
import socket
import ujson

from edge import eventlog
from edge.model.geolocation import Geolocation

from .types import EventArgs
from .stats import EventCounters
from .._models import Fact


class EventError(Exception):
    pass


class IEventService:
    def __init__(self):
        pass

    def consume_fact(self, fact_ref, fact):
        """
        :param fact_ref: (platear-id, media-id)
        :type fact_ref: (int, int)
        :type fact: Fact
        """
        raise NotImplementedError

    def destroy(self):
        pass


class _StubService(IEventService):
    def __init__(self):
        IEventService.__init__(self)

    def consume_fact(self, fact_ref, fact):
        """
        :param fact_ref: (platear-id, media-id)
        :type fact_ref: (int, int)
        :type fact: Fact
        """
        pass


class _Service(IEventService):
    def __init__(self, counters, args):
        """
        :type counters: EventCounters
        :type args: EventArgs
        """
        IEventService.__init__(self)
        self._mode = args.mode
        self._counters = counters
        self._udp_peer = (args.udp_peer_host, args.udp_peer_port)
        self._udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    def consume_fact(self, fact_ref, fact):
        """
        :param fact_ref: (platear-id, media-id)
        :type fact_ref: (int, int)
        :type fact: Fact
        """

        # <post-guid>,<post-channel>,<image-ts>,<plate-text>,<media-id>,
        # <speed>,<speed-limit>,<platear-id>,<geodata>,<fact-guid>

        if 'transit' != fact.jnode.get('type'):
            return

        def get_waypoints(jnode):
            """
            :param jnode: #/
            :type jnode: dict

            :return: #/measurements/0/waypoints
            :rtype: dict
            """
            m_jnode = jnode.get('measurements')
            if not m_jnode:
                eventlog.log_e_param('#/measurements')
                raise EventError

            w_jnode = m_jnode[0].get('waypoints')
            if not w_jnode:
                eventlog.log_e_param('#/measurements/0/waypoints')
                raise EventError

            return w_jnode

        def get_times(jnode):
            """
            :param jnode: #/measurements/waypoints
            :type jnode: dict

            :return: (init-ts, best-ts, fini-ts)
            :rtype: (int, int, int)
            """
            answer = list()
            for key in ['catch', 'best', 'lost']:
                dt = iso8601.parse_date(jnode[key]['time'])
                epoch_dt = datetime.datetime.fromtimestamp(0, dt.tzinfo)
                ts = 1e6 * (dt - epoch_dt).total_seconds()
                answer.append(int(ts))

            return answer

        def pack_geodata():
            """
            :rtype: str, dict
            """
            utf8, jnode = '-', dict()

            geo_jnode = fact.jnode['origin'].get('geolocation')
            if not geo_jnode:
                return utf8, jnode

            geo = Geolocation.deserialize(geo_jnode)
            if not geo:
                return utf8, jnode

            # <latChar><DDD.uuuuuu><longChar><ddd.uuuuuu>
            utf8 = '{}{:.6f}{}{:.6f}'.format(
                'N' if geo.latitude >= 0 else 'S', abs(geo.latitude),
                'E' if geo.longitude >= 0 else 'W', abs(geo.longitude))

            geo_jnode = {
                "latitude": geo.latitude,
                "longitude": geo.longitude
            }

            if geo.altitude is not None:
                geo_jnode['altitude'] = geo.altitude

            if geo.speed is not None:
                geo_jnode['speed'] = geo.speed

            return utf8, {'geodata': geo_jnode}

        waypoints_jnode = get_waypoints(fact.jnode)
        init_ts, best_ts, fini_ts = get_times(waypoints_jnode)
        geodata_utf8, geodata_jnode = pack_geodata()

        #
        fact_uuid = fact.jnode['uuid']

        best_waypoint = waypoints_jnode['best']
        vehicle_jnode = best_waypoint['vehicle']
        plate_jnode = None
        if len(vehicle_jnode.get('license-plates')):
            plate_jnode = vehicle_jnode['license-plates'][0]
            text = plate_jnode['text']['ucode']
            validity = plate_jnode['weight']
            plate_rect = plate_jnode['bbox-2']
            country = plate_jnode['type'].get('country', {}).get('numeric', 0)

        direction = {'approaching': 11, 'receding': 12, 'unknown': 1}.get(best_waypoint.get('dir', 'unknown'))

        lane = best_waypoint.get('location', {}).get('lane')
        if lane and len(lane) in [1, 2]:
            lane = int(lane)
        else:
            lane = None

        speed = vehicle_jnode.get('spot-speed', {}).get('radar', {}).get('measured', None)
        length = vehicle_jnode.get('size', {}).get('l', None)

        platear_id, media_id = fact_ref
        arg_list = [
            fact.jnode['origin'].get('uuid', '00000000-0000-0000-0000-000000000000'),
            0,
            best_ts,
            text,

            media_id,
            '-' if speed is None else speed,
            '-',
            platear_id,
            geodata_utf8,
            fact_uuid
        ]

        if self._mode == EventArgs.MODE_WITH_JSON:
            fact_jnode = {
                "id": platear_id,
                "uuid": fact_uuid,
                "catch-ts": init_ts,
                "lost-ts": fini_ts,
                "vehicle": {
                    "dir": direction
                },
                "image-ref": {
                    "ts": best_ts,
                    "id": media_id
                }
            }

            if length:
                fact_jnode['vehicle']['lenght'] = length

            if lane:
                fact_jnode['vehicle']['lane'] = lane

            if speed:
                fact_jnode['vehicle']['speed'] = speed

            if plate_jnode:
                plate_type = [country, 0, 0]

                fact_jnode['vehicle']['plate'] = {
                    "rect": plate_rect,
                    "type": plate_type,
                    "text": text,
                    "validity": validity
                }

            fact_jnode.update(geodata_jnode)

            fact_utf8 = ujson.dumps(fact_jnode, ensure_ascii=False)
            fact_base64 = base64.b64encode(fact_utf8)
            arg_list.append(fact_base64)

        datagram = u','.join(map(unicode, arg_list)).encode('utf-8')
        self._udp_socket.sendto(datagram, self._udp_peer)
        self._counters.inc_event_count()

    def destroy(self):
        self._udp_socket.close()


def init_service(counters, args):
    """
    :type counters: EventCounters
    :type args: EventArgs | None
    :rtype: IEventService
    """
    service = _Service(counters, args) if args else _StubService()
    return service


def fini_service(service):
    """
    :type service: IEventService | _Service | _StubService
    """
    service.destroy()
