import os

from edge import eventlog
from edge.com.db import DbArgs
from edge.types import JnodeError


class ApiArgs:
    def __init__(self, endpoint):
        """
        :type endpoint: str
        """
        self.endpoint = endpoint

    @staticmethod
    def unpack(jnode):
        """
        :param jnode: main.conf#/api
        :type jnode: str

        :rtype: ApiArgs
        """
        return ApiArgs(jnode)


class StoreArgs:
    def __init__(self, db, systest):
        """
        :type db: DbArgs
        :type systest: bool
        """
        self.db = db
        self.systest = systest

    @staticmethod
    def unpack(jnode):
        """
        :param jnode: main.conf#/
        :type jnode: dict

        :rtype: StoreArgs
        :raise JnodeError:
        """

        # unpack #/systest
        systest = jnode.get('systest', False)
        if type(systest) != bool:
            eventlog.log_e_param('#/systest')
            raise JnodeError

        # unpack #/db
        db_path = jnode.get('db')
        if not (type(db_path) in [unicode, str] and len(db_path)):
            eventlog.log_e_param('#/db')
            raise JnodeError

        try:
            db = DbArgs.deserialize2(db_path)
        except (KeyError, ValueError):
            eventlog.log_e_param('#/db')
            raise JnodeError

        return StoreArgs(db, systest)


class ReactArgs:
    def __init__(self, endpoints):
        """
        :type endpoints: list[str]
        """
        self.endpoints = endpoints

    @staticmethod
    def unpack(jnode):
        """
        :param jnode: main.conf#/fact-react
        :type jnode: list[str]

        :rtype: ReactArgs
        """
        return ReactArgs(jnode)


class EventArgs:
    MODE_DEFAULT = 1
    MODE_WITH_JSON = 2

    def __init__(self, mode, udp_peer_host, udp_peer_port):
        """
        :type mode: int
        :type udp_peer_host: str
        :type udp_peer_port: int
        """
        self.mode = mode
        self.udp_peer_host = udp_peer_host
        self.udp_peer_port = udp_peer_port

    @staticmethod
    def unpack(jnode):
        """
        :param jnode: main.conf#/fact-event
        :type jnode: (int, str, int)

        :rtype: EventArgs
        :raise JnodeError:
        """
        if len(jnode) != 3:
            eventlog.log_e_param('#/')
            raise JnodeError

        mode, udp_peer_host, udp_peer_port = jnode
        if mode not in [EventArgs.MODE_DEFAULT, EventArgs.MODE_WITH_JSON]:
            eventlog.log_e_param('#/0')
            raise JnodeError

        return EventArgs(mode, udp_peer_host, udp_peer_port)


class StatsArgs:
    def __init__(self, stat_filepath, stat_tout_sec):
        """
        :type stat_filepath: str
        :type stat_tout_sec: int
        """
        self.stat_filepath = stat_filepath
        self.stat_tout_sec = stat_tout_sec

    @staticmethod
    def unpack(jnode):
        """
        :param jnode: main.conf#/monitoring
        :type jnode: dict

        :rtype: StatsArgs
        :raise JnodeError:
        """

        # unpack #/stat-path
        stat_path = jnode.get('stat-path')
        if not (unicode == type(stat_path) and len(stat_path) < 64 and os.path.isabs(stat_path)):
            eventlog.log_e_param('#/stat-path')
            raise JnodeError

        # unpack #/stat-tout
        stat_tout_sec = jnode.get('stat-tout')
        if not (int == type(stat_tout_sec) and 0 <= stat_tout_sec <= 255):
            eventlog.log_e_param('#/stat-tout')
            raise JnodeError

        return StatsArgs(stat_path, stat_tout_sec)
