from edge.types import ImageRef


class JsonImRef(ImageRef):
    def __init__(self, *args, **kwargs):
        ImageRef.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(jnode):
        """
        :type jnode: dict
        :rtype: JsonImRef
        """
        blob_ts = jnode['Ts']
        blob_url = jnode.get('Url')
        if blob_url is None or len(blob_url) == 0:
            blob_url = None

        blob_format = jnode['Codec']
        blob_w, blob_h = jnode['Wh']

        return JsonImRef(blob_ts, blob_url, blob_format, int(blob_w), int(blob_h))


class RaMargin:
    def __init__(self, left, top, right, bottom):
        """
        :type left: float | int
        :type top: float | int
        :type right: float | int
        :type bottom: float | int
        """
        self.left = float(left)
        self.right = float(right)
        self.top = float(top)
        self.bottom = float(bottom)

    @staticmethod
    def unpack(jnode):
        """
        :param jnode: TrScene#/Ra-ltrb
        :type jnode: list[float]
        :rtype: RaMargin
        """
        if len(jnode) != 4:
            raise ValueError

        return RaMargin(*jnode)


class LrQuad:
    def __init__(self, x0, y0, x1, y1, x2, y2, x3, y3):
        """
        :type x0: int
        :type y0: int
        :type x1: int
        :type y1: int
        :type x2: int
        :type y2: int
        :type x3: int
        :type y3: int
        """

        self.x0 = x0
        self.y0 = y0

        self.x1 = x1
        self.y1 = y1

        self.x2 = x2
        self.y2 = y2

        self.x3 = x3
        self.y3 = y3

    def map(self, source_im_w, source_im_h, target_im_w, target_im_h, margin_tb=0, margin_lr=0):
        """
        :type source_im_w: int
        :type source_im_h: int
        :type target_im_w: int
        :type target_im_h: int
        :type margin_tb: int
        :type margin_lr: int

        :rtype: (int, int, int, int, int, int, int, int)
        """

        x0 = int(float(self.x0) / source_im_w * target_im_w - margin_lr)
        y0 = int(float(self.y0) / source_im_h * target_im_h - margin_tb)

        x1 = int(float(self.x1) / source_im_w * target_im_w + margin_lr)
        y1 = int(float(self.y1) / source_im_h * target_im_h - margin_tb)

        x2 = int(float(self.x2) / source_im_w * target_im_w + margin_lr)
        y2 = int(float(self.y2) / source_im_h * target_im_h + margin_tb)

        x3 = int(float(self.x3) / source_im_w * target_im_w - margin_lr)
        y3 = int(float(self.y3) / source_im_h * target_im_h + margin_tb)

        return x0, y0, x1, y1, x2, y2, x3, y3

    def get_rel_bbox(self, source_im_w, source_im_h):
        """
        :type source_im_w: int
        :type source_im_h: int
        :rtype: (float, float, float, float)
        """
        fl = float(min(self.x0, self.x3)) / source_im_w
        fr = float(max(self.x1, self.x2)) / source_im_w
        ft = float(min(self.y0, self.y1)) / source_im_h
        fb = float(max(self.y2, self.y3)) / source_im_h

        return fl, ft, fr, fb

    def map_bbox_wh(self, source_im_w, source_im_h, target_im_w, target_im_h):
        """
        :type source_im_w: int
        :type source_im_h: int
        :type target_im_w: int
        :type target_im_h: int
        :rtype: (int, int)
        """
        bbox_il = min(self.x0, self.x3)
        bbox_ir = max(self.x1, self.x2)
        bbox_it = min(self.y0, self.y1)
        bbox_ib = max(self.y2, self.y3)

        target_bbox_fw = float(bbox_ir - bbox_il) / source_im_w * target_im_w
        target_bbox_fh = float(bbox_ib - bbox_it) / source_im_h * target_im_h
        return int(target_bbox_fw), int(target_bbox_fh)

    def map_bbox_ref(self, source_im_w, source_im_h, target_im_w, target_im_h):
        """
        :type source_im_w: int
        :type source_im_h: int
        :type target_im_w: int
        :type target_im_h: int
        :rtype: (int, int)
        """
        target_bbox_il = float(min(self.x0, self.x3)) / source_im_w * target_im_w
        target_bbox_ir = float(max(self.x1, self.x2)) / source_im_w * target_im_w
        target_bbox_fb = float(max(self.y2, self.y3)) / source_im_h * target_im_h

        return int(0.5 * (target_bbox_il + target_bbox_ir)), int(target_bbox_fb)

    def map_bbox_ref2(self, source_im_w, source_im_h, target_im_w, target_im_h):
        """
        :type source_im_w: int
        :type source_im_h: int
        :type target_im_w: int
        :type target_im_h: int
        :rtype: (int, int)
        """
        target_bbox_il = float(min(self.x0, self.x3)) / source_im_w * target_im_w
        target_bbox_ir = float(max(self.x1, self.x2)) / source_im_w * target_im_w
        target_bbox_ft = float(max(self.y0, self.y1)) / source_im_h * target_im_h

        return int(0.5 * (target_bbox_il + target_bbox_ir)), int(target_bbox_ft)

    @staticmethod
    def unpack(jnode):
        """
        :type jnode: list[float]
        :rtype: LrQuad
        """
        if len(jnode) != 8:
            raise RuntimeError

        return LrQuad(*jnode)


class FaceBbox:
    def __init__(self, x0, y0, x1, y1):
        """
        :type x0: int
        :type y0: int
        :type x1: int
        :type y1: int
        """

        self.x0 = x0
        self.y0 = y0

        self.x1 = x1
        self.y1 = y1

    @staticmethod
    def unpack(jnode):
        """
        :type jnode: list[float]
        :rtype: FaceBbox
        """
        if len(jnode) != 4:
            raise RuntimeError

        return FaceBbox(*jnode)


class LrSceneItem:
    def __init__(self, lr_oid, lr_dir, lr_text, lr_quad, lr_trace):
        """
        :type lr_oid: int
        :type lr_dir: int
        :type lr_text: str
        :type lr_quad: LrQuad
        :type lr_trace: list

        """
        self.lr_oid = lr_oid
        self.lr_dir = lr_dir
        self.lr_text = lr_text
        self.lr_quad = lr_quad
        self.lr_trace = lr_trace

    def map_bbox_ref(self, source_im_w, source_im_h, target_im_w, target_im_h):
        """
        :type source_im_w: int
        :type source_im_h: int
        :type target_im_w: int
        :type target_im_h: int
        :rtype: (int, int)
        """
        return self.lr_quad.map_bbox_ref(source_im_w, source_im_h, target_im_w, target_im_h)

    def map_bbox_ref2(self, source_im_w, source_im_h, target_im_w, target_im_h):
        """
        :type source_im_w: int
        :type source_im_h: int
        :type target_im_w: int
        :type target_im_h: int
        :rtype: (int, int)
        """
        return self.lr_quad.map_bbox_ref2(source_im_w, source_im_h, target_im_w, target_im_h)

    @staticmethod
    def unpack(jnode):
        """
        :param jnode: TrScene#/Lr/:index
        :param jnode: TrScene#/LrRa/:index/Lr
        :type jnode: dict
        :rtype: LrSceneItem
        """
        lr_plate_jnode = jnode['Plate']

        lr_oid = jnode['Id']
        lr_dir = jnode['Dir']
        lr_text = lr_plate_jnode['P']
        lr_quad = LrQuad.unpack(lr_plate_jnode['Q'])
        lr_trace = jnode['Trace']

        return LrSceneItem(lr_oid, lr_dir, lr_text, lr_quad, lr_trace)


class FaceSceneItem:
    def __init__(self, face_id, face_conf, face_bbox):
        """
        :type face_id: int
        :type face_conf: float
        :type face_bbox: FaceBbox
        """
        self.face_id = face_id
        self.face_conf = face_conf
        self.face_bbox = face_bbox

    @staticmethod
    def unpack(jnode):
        """
        :param jnode: TrScene#/Lr/:index
        :param jnode: TrScene#/LrRa/:index/Lr
        :type jnode: dict
        :rtype: FaceSceneItem
        """
        face_id = jnode['Id']
        face_conf = jnode['Conf']
        face_bbox = FaceBbox.unpack(jnode['R'])

        return FaceSceneItem(face_id, face_conf, face_bbox)


class RaSceneItem:
    def __init__(self, rel_im_x, rel_im_y, ra_speed, ra_len, ra_dir, ra_ltrb):
        """
        :type rel_im_x: float
        :type rel_im_y: float
        :type ra_speed: int
        :type ra_len: int
        :type ra_dir: int
        :type ra_ltrb: RaMargin
        """
        self.ra_rel_im_x = rel_im_x
        self.ra_rel_im_y = rel_im_y

        self.ra_speed = ra_speed
        self.ra_len = ra_len
        self.ra_dir = ra_dir
        self.ra_ltrb = ra_ltrb

    def map_bbox_ref(self, target_im_w, target_im_h):
        """
        :type target_im_w: int
        :type target_im_h: int
        :rtype: (int, int)
        """
        abs_x = self.ra_rel_im_x * target_im_w
        abs_y = self.ra_rel_im_y * target_im_h
        return int(abs_x), int(abs_y)

    @staticmethod
    def unpack(jnode):
        """
        :param jnode: TrScene#/Ra/:index
        :param jnode: TrScene#/LrRa/:index/Ra
        :type jnode: dict
        :rtype: RaSceneItem
        """
        rel_im_x, rel_im_y = jnode['xy']
        ra_speed = int(jnode['speed'])
        ra_len = int(jnode['len'])
        ra_dir = int(jnode['dir'])
        ra_ltrb = RaMargin.unpack(jnode['r'])

        return RaSceneItem(float(rel_im_x), float(rel_im_y), ra_speed, ra_len, ra_dir, ra_ltrb)


class TrSceneItem:
    def __init__(self, lr_item, ra_item):
        """
        :type lr_item: LrSceneItem
        :type ra_item: RaSceneItem
        """
        self.lr_item = lr_item
        self.ra_item = ra_item

    @staticmethod
    def unpack(jnode):
        """
        :param jnode: TrScene#/LrRa/:index
        :type jnode: dict
        :rtype: TrSceneItem
        """
        lr_item = LrSceneItem.unpack(jnode['Lr'])
        ra_item = RaSceneItem.unpack(jnode['Ra'])
        return TrSceneItem(lr_item, ra_item)


class TrScene:
    def __init__(self, im_ref, tr_items, lr_items, ra_items, face_items):
        """
        :type im_ref: ImageRef
        :type tr_items: list[TrSceneItem]
        :type lr_items: list[LrSceneItem]
        :type ra_items: list[RaSceneItem]
        :type face_items: list[FaceSceneItem]
        """
        self.im_ref = im_ref

        self.tr_items = tr_items
        self.lr_items = lr_items
        self.ra_items = ra_items
        self.face_items = face_items

    @staticmethod
    def unpack(jnode):
        """
        :param jnode: TrScene#/
        :type jnode: dict
        :rtype: TrScene
        """
        im_ref = JsonImRef.deserialize(jnode['Image'])

        tr_items = map(lambda item_jnode: TrSceneItem.unpack(item_jnode), jnode.get('LrRa', []))
        lr_items = map(lambda item_jnode: LrSceneItem.unpack(item_jnode), jnode.get('Lr', []))
        ra_items = map(lambda item_jnode: RaSceneItem.unpack(item_jnode), jnode.get('Ra', []))
        face_items = map(lambda item_jnode: FaceSceneItem.unpack(item_jnode), jnode.get('Face', []))

        return TrScene(im_ref, tr_items, lr_items, ra_items, face_items)
