import ujson
from os.path import join as join_path
from os.path import exists as exists_path

from edge import eventlog
from edge.com import fs

from edge.types import JnodeError
from edge.service.types import TzServiceArgs, AboutServiceArgs
from edge.service.jnode.types import JsonGpsdServiceArgs, JsonTzServiceArgs

from edge.landmark.jnode.types import LandmarkServiceArgs, JsonLandmarkServiceArgs

from .detectors.base.jnode.types import JsonDetectorArgs as JsonBaseArgs

from .detectors.transit.types import DetectorArgs as TransitArgs
from .detectors.transit.jnode.types import JsonDetectorArgs as JsonTransitArgs

from .detectors.parking.types import DetectorArgs as ParkingArgs
from .detectors.parking.jnode.types import JsonDetectorArgs as JsonParkingArgs

from .detectors.wrong_way.types import DetectorArgs as WrongWayArgs
from .detectors.wrong_way.jnode.types import JsonDetectorArgs as JsonWrongWayArgs

from .detectors.bus_lane.types import DetectorArgs as BusLaneArgs
from .detectors.bus_lane.jnode.types import JsonDetectorArgs as JsonBusLaneArgs

from .detectors.spot_speed.types import DetectorArgs as SpotSpeedArgs
from .detectors.spot_speed.jnode.types import JsonDetectorArgs as JsonSpotSpeedArgs

from .detectors.red_light.types import DetectorArgs as RedLightArgs
from .detectors.red_light.jnode.types import JsonDetectorArgs as JsonRedLightArgs

# from .detectors.red_light_leddar.types import DetectorArgs as RedLightArgs2
from .detectors.red_light_leddar.jnode.types import JsonDetectorArgs as JsonRedLightArgs2

from .detectors.section_speed.types import DetectorArgs as SectionSpeedArgs
from .detectors.section_speed.jnode.types import JsonDetectorArgs as JsonSectionSpeedArgs

from .detectors.wanted.types import DetectorArgs as WantedArgs
from .detectors.wanted.jnode.types import JsonDetectorArgs as JsonWantedArgs

from .services.types import (RaasServiceArgs, TrainerServiceArgs,
                             StatServiceArgs, LaneDetectorServiceArgs,
                             VsServiceArgs, DbServiceArgs, SignServiceArgs)
from .services.jnode.types import (JsonBsServiceArgs, JsonRaasServiceArgs,
                                   JsonReactServiceArgs, JsonEventServiceArgs,
                                   JsonTrainerServiceArgs, JsonStatServiceArgs,
                                   JsonLaneDetectorServiceArgs,
                                   JsonTlServiceArgs, TlServiceArgs,
                                   JsonDbServiceArgs, JsonSystestServiceArgs,
                                   JsonSignServiceArgs, JsonLeddarServiceArgs,
                                   JsonPlateInfoServiceConf)


class DetectorsArgs:
    def __init__(self, parking, wrong_way, spot_speed, red_light, wanted,
                 section_speed, railroad_crossing, bus_lane):
        """
        :type parking: list[ParkingArgs]
        :type wrong_way: list[WrongWayArgs]
        :type spot_speed: list[SpotSpeedArgs]
        :type red_light: list[RedLightArgs]
        :type wanted: list[WantedArgs]
        :type section_speed: list[SectionSpeedArgs]
        :type railroad_crossing: list[JsonRedLightArgs2]
        :type bus_lane: list[BusLaneArgs]
        """
        self.parking = parking
        self.wrong_way = wrong_way
        self.spot_speed = spot_speed
        self.red_light = red_light
        self.section_speed = section_speed
        self.wanted = wanted
        self.railroad_crossing = railroad_crossing
        self.bus_lane = bus_lane

    @staticmethod
    def deserialize(detectors_jnode, def_footer_mapping):
        """
        :type detectors_jnode: dict
        :type def_footer_mapping: dict
        :rtype: DetectorsArgs
        """
        def deserialize_detectors(conf_key, parser_fn):
            """
            :type conf_key: str
            :type parser_fn: (dict, int, dict) -> T
            :rtype: list of T
            """
            conf_path_list = detectors_jnode.get(conf_key, [])
            if type(conf_path_list) != list or len(conf_path_list) > 64:
                eventlog.log_e_param('#/detectors/%s' % conf_key)
                raise JnodeError

            ident = 0
            args_list = []
            for conf_path in conf_path_list:
                with open(conf_path) as json_file:
                    detector_jnode = ujson.load(json_file)

                args_list.append(parser_fn(detector_jnode, ident, def_footer_mapping))
                ident += 1

            return args_list

        parking = deserialize_detectors('parking.conf', JsonParkingArgs.deserialize)
        wrong_way = deserialize_detectors('wrong-way.conf', JsonWrongWayArgs.deserialize)
        bus_lane = deserialize_detectors('bus-lane.conf', JsonBusLaneArgs.deserialize)
        spot_speed = deserialize_detectors('spot-speed.conf', JsonSpotSpeedArgs.deserialize)
        red_light = deserialize_detectors('red-light.conf', JsonRedLightArgs.deserialize)
        red_light2 = deserialize_detectors('railroad-crossing.conf', JsonRedLightArgs2.deserialize)
        wanted = deserialize_detectors('wanted.conf', JsonWantedArgs.deserialize)
        section_speed = deserialize_detectors('section-speed.conf',
                                              JsonSectionSpeedArgs.deserialize)

        return DetectorsArgs(parking, wrong_way, spot_speed,
                             red_light, wanted, section_speed, red_light2, bus_lane)


class KernelArgs:
    def __init__(self, lane_detector_args, trainer, react_service, event_consumer,
                 counters, tz, raas, tl, gpsd, about, base_detector, detectors, blob_store, video_store,
                 sign_args, transit_args, landmark_args, db_args, systest, ledar, plate_info,
                 use_videostore):
        """
        :type lane_detector_args: LaneDetectorServiceArgs
        :type trainer: TrainerServiceArgs
        :type react_service: JsonReactServiceArgs
        :type event_consumer: JsonEventServiceArgs
        :type counters: StatServiceArgs
        :type tz: TzServiceArgs
        :type raas: RaasServiceArgs | None
        :type tl: TlServiceArgs | None
        :type gpsd: JsonGpsdServiceArgs
        :type about: AboutServiceArgs
        :type base_detector: JsonBaseArgs
        :type detectors: DetectorsArgs
        :type blob_store: JsonBsServiceArgs
        :type video_store: VsServiceArgs
        :type sign_args: SignServiceArgs | None
        :type transit_args : TransitArgs
        :type landmark_args: LandmarkServiceArgs | None
        :type db_args: DbServiceArgs
        :type systest: JsonSystestServiceArgs
        :type ledar: JsonLeddarServiceArgs | None
        :type plate_info: JsonPlateInfoServiceConf
        :type use_videostore: bool
        """
        self.lane_detector = lane_detector_args
        self.trainer = trainer

        self.react_service = react_service
        self.event_service = event_consumer
        self.db = db_args

        self.counters = counters

        self.tz = tz
        self.raas = raas
        self.tl = tl
        self.ledar = ledar
        self.gpsd = gpsd
        self.about = about

        self.detectors = detectors
        self.transit_args = transit_args

        self.base_detector = base_detector
        self.bs = blob_store
        self.vs = video_store
        self.sign = sign_args
        self.systest = systest
        self.plate_info = plate_info
        self.use_videostore = use_videostore
        self.landmark_args = landmark_args

    @staticmethod
    def deserialize(main_conf_jnode, log_level, module_id):
        """
        :type main_conf_jnode: dict
        :type log_level: str
        :type module_id: int
        :rtype: KernelArgs
        """
        # define blobstore args
        tmp_path = join_path('/run', 'edge', 'sc', str(module_id))
        video_store = VsServiceArgs(module_id, tmp_path, log_level=log_level)

        # define services
        services_jnode = main_conf_jnode.get('services')
        if type(services_jnode) != dict:
            eventlog.log_e_param('#/services')
            raise JnodeError

        # define stat-counters-args
        monitoring_jnode = services_jnode.get('stat')
        if not monitoring_jnode:
            raise ValueError('#/services/stat')

        use_videostore = services_jnode.get('vs', True)

        stat_service = JsonStatServiceArgs.deserialize(monitoring_jnode)
        systest = JsonSystestServiceArgs.deserialize(main_conf_jnode)
        template = JsonPlateInfoServiceConf.deserialize(main_conf_jnode)

        # define def-footer-mapping
        tz_conf_filepath = services_jnode['tz.conf']
        with open(tz_conf_filepath) as tz_conf_jnode:
            tz_conf_jnode = ujson.load(tz_conf_jnode)

        tz = JsonTzServiceArgs.deserialize(tz_conf_jnode)

        # define road args
        road_conf_jnode = {}
        road_conf_filepath = services_jnode.get('road.conf')
        if road_conf_filepath:
            with open(road_conf_filepath) as jnode:
                road_conf_jnode = ujson.load(jnode)

        dir_jnode = road_conf_jnode.get('road', dict()).get('dir') or dict()

        lane_detector_args = JsonLaneDetectorServiceArgs.deserialize(road_conf_jnode)
        # define origin args
        conf_path = services_jnode.get('origin')
        about = AboutServiceArgs(conf_path)

        gps_conf_filepath = services_jnode.get('gpsd')
        with open(gps_conf_filepath) as geolocation_jnode:
            geolocation_jnode = ujson.load(geolocation_jnode)

        gpsd = JsonGpsdServiceArgs.deserialzie(geolocation_jnode)

        def_footer_mapping = {
            'tz': tz.tz_info,
            'post-name': '',
            'geolocation': '',
            'plate': '',
            'speed': '-',
            'approaching': dir_jnode.get('approaching') or 'approaching',
            'receding': dir_jnode.get('receding') or 'receding',
            'unknown': dir_jnode.get('unknown') or 'unknown',
            'video-speed': '-',
            'video-speed2': '-',
            'radar-speed': '-',
            'max-speed': '-',
            'speed-limit': '-',
            'radar-overspeed': '-'
        }
        # define db args
        db_conf_abspath = services_jnode.get('db.conf')
        if not db_conf_abspath:
            eventlog.log_e('fail: param #/services/db.conf has invalid value')
            raise JnodeError

        try:
            with open(db_conf_abspath) as db_conf:
                db_conf_jnode = ujson.load(db_conf)

        except Exception as x:
            eventlog.log_e('fail: load %s (%s)' % (db_conf_abspath, x))
            raise JnodeError

        db_args = JsonDbServiceArgs.deserialize(module_id, db_conf_jnode)

        # define detectors args
        detectors_jnode = main_conf_jnode.get('detectors', {})
        detectors = DetectorsArgs.deserialize(detectors_jnode, def_footer_mapping)

        # define transit args
        transit_conf_list = detectors_jnode.get('transit.conf')
        if type(transit_conf_list) != list or len(transit_conf_list) != 1:
            eventlog.log_e('fail: #/detectors/transit.conf has invalid value')
            raise JnodeError

        with open(transit_conf_list[0]) as jnode:
            transit_conf_jnode = ujson.load(jnode)

        transit_args = JsonTransitArgs.deserialize(transit_conf_jnode)
        base_detector = JsonBaseArgs.deserialize(transit_conf_jnode, def_footer_mapping)

        # define blob-store-args
        blob_store = JsonBsServiceArgs.deserialize(
            log_level, transit_args,
            detectors.wrong_way,
            detectors.spot_speed,
            detectors.red_light,
            detectors.railroad_crossing,
            detectors.parking,
            detectors.wanted,
            detectors.section_speed,
            detectors.bus_lane
        )

        # define others
        trainer_jnode = services_jnode.get('trainer')
        trainer = JsonTrainerServiceArgs.deserialize(trainer_jnode) if trainer_jnode else None

        react_service = JsonReactServiceArgs.deserialize(main_conf_jnode)

        # define event-consumer-args
        event_service = JsonEventServiceArgs.deserialize(services_jnode['event'])

        # deserialize #/raas
        raas = JsonRaasServiceArgs.deserialize(services_jnode.get('raas', None))

        # define traffic-ligth service args
        tl = JsonTlServiceArgs.deserialize(services_jnode.get('tl', None))

        # define leddar service args
        leddar = JsonLeddarServiceArgs.deserialize(services_jnode.get('leddar', None))

        # define sign service args
        sign_args = None
        sign_jnode = services_jnode.get('sign')
        if sign_jnode:
            sign_args = JsonSignServiceArgs.deserialize(sign_jnode, log_level)

        landmark_args = None
        for landmark_path in ['/etc/edge/radar.0/runtime/location.conf', services_jnode.get('image-mapper.conf')]:
            if landmark_path and exists_path(landmark_path) and (landmark_args is None):
                ra_conf_jnode = fs.load_json3(landmark_path, default=None)
                if ra_conf_jnode is not None:
                    landmark_args = JsonLandmarkServiceArgs.deserialize(ra_conf_jnode)

        kernel_args = KernelArgs(
            lane_detector_args, trainer, react_service, event_service,
            stat_service, tz, raas, tl, gpsd, about, base_detector, detectors, blob_store, video_store,
            sign_args, transit_args, landmark_args, db_args, systest, leddar, template, use_videostore)

        return kernel_args
