from edge_bbox.types import RectF, QuadI
from edge.model.track import LrTrackItem
from ...model.models import RaTrackItem, TrTrackItem

_FAKE_PLATE_SIZE = 0.01


def check_quad_dist_from_img_borders(plate_quad, lr_scene_w, lr_scene_h):
    """
    :type plate_quad: QuadI
    :type lr_scene_w: int
    :type lr_scene_h: int
    :rtype: bool
    """
    bbox = RectF(*plate_quad.get_rel_bbox(lr_scene_w, lr_scene_h))

    plate_w = bbox.fr - bbox.fl
    plate_h = bbox.fb - bbox.ft

    return (bbox.ft > plate_h and
            bbox.fb < (1.0 - 2 * plate_h) and
            bbox.fl > plate_w and
            bbox.fr < (1.0 - plate_w))


def get_lr_item_dist(lr_item, lr_scene_w, lr_scene_h):
    """
    :type lr_item: LrTrackItem
    :type lr_scene_w: int
    :type lr_scene_h: int
    :rtype: long
    """
    bbox = RectF(*lr_item.plate_quad.get_rel_bbox(lr_scene_w, lr_scene_h))

    # vehicle's plate is fully visible on the image (1 bit).
    dist1 = 1 if bbox.get_visible_100() else 0

    # the weight of plate's recognition (7 bits).
    dist2 = lr_item.plate_weight

    # vehicle's body is more visible (1 bit).
    dist3 = 1 if check_quad_dist_from_img_borders(lr_item.plate_quad, lr_scene_w, lr_scene_h) else 0

    # plate's height [0, 1000] (10 bits).
    dist4 = int((bbox.fb - bbox.ft) * 1000)

    dist = (dist1 << 18) + (dist2 << 11) + (dist3 << 10) + dist4
    return dist


def get_ra_item_dist(ra_item, lr_item=None, lr_scene_w=None, lr_scene_h=None):
    """
    :type ra_item: RaTrackItem
    :type lr_item: LrTrackItem | None
    :type lr_scene_w: int | None
    :type lr_scene_h: int | None
    :rtype: int
    """
    # if None in [lr_item, lr_scene_w, lr_scene_h]:
    #     best_line_y = 0.5
    # else:
    #     l, t, r, b = lr_item.plate_quad.get_rel_bbox(lr_scene_w, lr_scene_h)
    #     best_line_y = (t + b) / 2.0

    # ra-item is fully visible on the image (1 bit).
    def check_borders(coord):
        return (0.0 + _FAKE_PLATE_SIZE) <= coord <= (1.0 - _FAKE_PLATE_SIZE)

    dist1 = 1 if check_borders(ra_item.x) and check_borders(ra_item.y) else 0

    # radar item is closer to the center of number plate (7 bit)

    #   (doesn't work because the farther from camera the car is locate,
    #   the less is the distance between ra-point and center of number plate

    # dist2 = 100 - abs(int(100 * (ra_item.y - best_line_y)))

    dist2 = 0

    dist = (dist1 << 7) + dist2
    return dist


def get_tr_item_dist(tr_item, lr_scene_w, lr_scene_h):
    """
    :type tr_item: TrTrackItem
    :type lr_scene_w: int | None
    :type lr_scene_h: int | None
    :rtype: int
    """
    lr_dist = 0
    if tr_item.lr_item:
        # ? to 468 968
        lr_dist = get_lr_item_dist(tr_item.lr_item, lr_scene_w, lr_scene_h)

    ra_dist = 0
    if tr_item.ra_item:
        # 1 to 228 when ra item on image
        # can be negative when ra item out off image
        ra_dist = get_ra_item_dist(tr_item.ra_item, tr_item.lr_item, lr_scene_w, lr_scene_h)

    return lr_dist + ra_dist * 10


def get_best_ra_ts_and_rect(track_items, search_image_list):
    """
    :type track_items: list of RaTrackItem
    :type search_image_list: list of int
    :rtype: (int, RectF)| None
    """
    best_line_y = 0.5
    best_item_dist = 2.0
    best_item_index = None
    for index, item in enumerate(track_items):
        dist = abs(item.y - best_line_y)
        if dist < best_item_dist:
            if item.ts in search_image_list:
                # noinspection PyChainedComparisons
                if item.x < 1.0 - _FAKE_PLATE_SIZE and item.x > _FAKE_PLATE_SIZE \
                        and item.y < 1.0 - _FAKE_PLATE_SIZE and item.y > _FAKE_PLATE_SIZE:
                    best_item_dist = dist
                    best_item_index = index

    if best_item_index is None:
        return None

    radar_point = track_items[best_item_index]

    plate_rect = RectF(
        radar_point.x - _FAKE_PLATE_SIZE, radar_point.y - _FAKE_PLATE_SIZE,
        radar_point.x + _FAKE_PLATE_SIZE, radar_point.y + _FAKE_PLATE_SIZE)

    return track_items[best_item_index].ts, plate_rect
