from edge import eventlog
from edge.service.about import IAboutService
from edge.service import db_wanted

from .types import DetectorArgs, WantedEvidence, RuleArgs, LimitsArgs
from ..types import IDetector
from ...model.types import BaseEvent, Event, BestWaypoint


from ...model.models import Track


def create(*args, **kwargs):
    """
    :rtype: IDetector
    """
    return _Detector(*args, **kwargs)


class _RuleServices:
    def __init__(self, about_service, wanted_service):
        """
        :type wanted_service: db_wanted.DbWantedApi
        :type about_service: IAboutService
        """
        self.wanted_service = wanted_service
        self.about = about_service


class Rule:
    def __init__(self, services, rule_args, detector_name):
        """
        :type services: _RuleServices
        :type rule_args: RuleArgs
        :type detector_name: str
        :rtype: WantedEvidence | None
        """
        self._detector_name = detector_name

        self._services = services
        self._args = rule_args

    def check(self, base):
        """
        :type base: BaseEvent
        :rtype: Deferred | None
        """
        all_points = base.measurements[0].waypoints
        best_points = filter(
            lambda waypoint: waypoint.name == waypoint.NAME_BEST,
            all_points)
        if len(best_points) != 1:
            eventlog.log_e('fail: find best-waypoint')
            return

        best_point = best_points[0]  # type: BestWaypoint
        with self._services.wanted_service as api:
            matches = api.check_plate(
                best_point.vehicle.plate_text,
                self._args.limits.plate_dist)

        if len(matches) == 0:
            return

        # filter matches
        rule_limits = self._args.limits

        rule_matches = []
        for match in matches:
            owner = match.list.owner
            name = match.list.ident

            if rule_limits.list_owner == '*' or rule_limits.list_owner == owner:
                if rule_limits.list_name == '*' or rule_limits.list_name == name:
                    rule_matches.append(match)

        if len(rule_matches) == 0:
            return

        # eventlog.log_i('get wanted plate %s' % best_waypoint.vehicle.plate_text)
        origin = self._services.about.get_origin(best_point.ts)
        evidence = WantedEvidence(origin, best_point.ts, self._args.name, self._args, rule_matches)
        return evidence


class _Detector(IDetector):
    def __init__(self, about_service, wanted_service, args):
        """
        :type about_service: IAboutService
        :type wanted_service: db_wanted.DbWantedApi
        :type args: DetectorArgs
        """
        self._args = args
        self._identity = args.identity
        self._handler = None

        self._rules = []
        rule_services = _RuleServices(about_service, wanted_service)
        for rule in args.rules:
            if rule.enabled:
                self._rules.append(Rule(rule_services, rule, args.name))

        IDetector.__init__(self)

    def set_handler(self, event_handler):
        """
        :type event_handler: (Event) -> None
        """
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler

    def analyze(self, track, base):
        """
        :type track: Track
        :type base: BaseEvent
        """

        for rule in self._rules:
            evidence = rule.check(base)
            if evidence:
                self._handler(Event(self._identity, base, evidence))
