from edge import eventlog
from edge.image.types import Blob, FileBlob, MemBlob, StructBlobMeta

from edge.com import bs

from .types import BsServiceArgs
from ..model.types import Identity


class IBsService:
    def __init__(self):
        pass

    def get_size(self, sender, blob_id, profile):
        """
        :type sender: Identity
        :type profile: str
        :type blob_id: str
        :rtype: int
        """
        raise NotImplementedError

    def save(self, sender, blob):
        """
        :type sender: Identity
        :type blob: Blob
        :raise bs.BsError:
        """
        raise NotImplementedError


class _BsService(IBsService):
    def __init__(self, twisted_reactor, bs_conf, counters):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorUNIX
        :type bs_conf: BsServiceArgs
        :type counters: bs.IBsCounters
        """
        IBsService.__init__(self)
        self._reactor = twisted_reactor
        self._session_args = counters, 65535, bs.LOG_LEVEL_NAME_TO_CODE_MAP.get(bs_conf.log_level, 'info')
        self._session = bs.Session(counters=counters, closed=True)

        self._counters = counters
        self._sender_to_chan_map = bs_conf.sender_to_chan_map

    def destroy(self):
        self._session.close()

    def get_size(self, sender, blob_id, profile):
        """
        :type sender: Identity
        :type profile: str
        :type blob_id: str
        :rtype: int
        """
        chan_or_profiles = self._sender_to_chan_map[sender]
        if dict == type(chan_or_profiles):
            chan = chan_or_profiles[profile]
        else:
            chan = chan_or_profiles

        reply = self._session.get_blob_size(chan, blob_id)
        return reply

    def save(self, sender, blob):
        """
        :type sender: Identity
        :type blob: MemBlob | FileBlob
        """
        chan_or_profiles = self._sender_to_chan_map[sender]
        if dict == type(chan_or_profiles):
            chan = chan_or_profiles[blob.profile]
        else:
            chan = chan_or_profiles

        self._counters.inc_blob(chan)
        self._save(chan, blob)

    @bs.reopen_session
    def _save(self, chan, blob):
        """
        :type chan: str
        :type blob: MemBlob | FileBlob

        :raise bs.BsError:
        :raise bs.BsSessionError:
        """
        try:
            blob_ts = blob.min_ts / 1000
            if blob.file_path is None:
                if isinstance(blob.meta, StructBlobMeta):
                    ident = self._session.save_blob(chan, blob_ts, blob.file_data, blob.meta.struct)
                else:
                    ident = self._session.save_blob(chan, blob_ts, blob.file_data, blob.file_meta)
            else:
                ident = self._session.save_blob2(chan, blob_ts, blob.file_path, blob.file_meta)

        except bs.BsError:
            eventlog.log_e('fail: save <chan={}, profile={}>'.format(str(chan), blob.profile))
            raise

        blob.name = '%s_%s.%s' % (chan, ident, blob.file_ext)


def init_service(twisted_reactor, args, counters):
    """
    :type twisted_reactor: twisted.internet.interfaces.IReactorUNIX
    :type args: BsServiceArgs
    :type counters: bs.IBsCounters
    :rtype: IBsService
    """
    return _BsService(twisted_reactor, args, counters)


def fini_service(service):
    """
    :type service: IBsService | _BsService
    """
    service.destroy()
