import sortedcontainers
import time

from edge import eventlog


def init_service(twisted_reactor):
    """
    :type twisted_reactor: twisted.internet.interfaces.IReactorUNIX
    :rtype: IClockService
    """
    return _ClockService(twisted_reactor)


def fini_service(service):
    """
    :type service: IClockService | _ClockService
    """
    service.destroy()


class _ControlTsStore:
    def __init__(self, max_entry_count, max_age_delta_usec, max_lookup_dist_usec):
        """
        :type max_entry_count: int
        :type max_age_delta_usec: long
        :type max_lookup_dist_usec: long
        """
        self._max_entry_count = max_entry_count
        self._max_age_delta_usec = max_age_delta_usec
        self._max_lookup_dist_usec = max_lookup_dist_usec

        self._ts_to_control_ts_map = sortedcontainers.SortedDict()

    def add_entry(self, ts, control_ts):
        """
        :type ts: long
        :type control_ts: long
        """
        self._ts_to_control_ts_map[ts] = control_ts

        while len(self._ts_to_control_ts_map) > self._max_entry_count:
            self._ts_to_control_ts_map.popitem(last=False)

        while (self._ts_to_control_ts_map.peekitem()[0] - self._ts_to_control_ts_map.peekitem(0)[0]) \
                > self._max_age_delta_usec:
            self._ts_to_control_ts_map.popitem(last=False)

    def lookup_nearest(self, lookup_ts):
        """
        :type lookup_ts: long
        :rtype: long | None
        """
        entry_count = len(self._ts_to_control_ts_map)
        if not entry_count:
            return None

        # insert lookup_ts
        bisect_index = self._ts_to_control_ts_map.bisect_right(lookup_ts)
        if 0 < bisect_index < entry_count:
            # exact-match with lookup_ts
            less_equal_ts_dist = lookup_ts - self._ts_to_control_ts_map.iloc[bisect_index - 1]
            if less_equal_ts_dist == 0:
                return self._ts_to_control_ts_map.peekitem(bisect_index - 1)[1]

            # greater_ts is better
            greater_ts_dist = self._ts_to_control_ts_map.iloc[bisect_index] - lookup_ts
            if less_equal_ts_dist < greater_ts_dist and less_equal_ts_dist <= self._max_lookup_dist_usec:
                return self._ts_to_control_ts_map.peekitem(bisect_index - 1)[1]

            # less_equal_ts is better
            return self._ts_to_control_ts_map.peekitem(bisect_index)[1] \
                if greater_ts_dist <= self._max_lookup_dist_usec else None

        # push-back lookup_ts
        if bisect_index == entry_count:
            less_equal_ts_dist = lookup_ts - self._ts_to_control_ts_map.iloc[entry_count - 1]
            return self._ts_to_control_ts_map.peekitem(bisect_index - 1)[1] \
                if less_equal_ts_dist <= self._max_lookup_dist_usec else None

        # push-front lookup_ts
        greater_ts_dist = self._ts_to_control_ts_map.iloc[bisect_index] - lookup_ts
        if greater_ts_dist <= self._max_lookup_dist_usec:
            return self._ts_to_control_ts_map.peekitem(bisect_index)[1]
        else:
            eventlog.log_d("fail: target-ts:{}usec, min-ts:{}usec, dt:{}sec".format(
                lookup_ts, self._ts_to_control_ts_map.peekitem(0)[0],
                (self._ts_to_control_ts_map.peekitem(0)[0] - lookup_ts) / 1e6))
            return None

    def get_count(self):
        return len(self._ts_to_control_ts_map)


class IClockService:
    def __init__(self):
        pass

    def get_control_ts(self, system_ts):
        """
        :type system_ts: long
        :rtype: long | None
        """
        raise NotImplementedError

    def destroy(self):
        raise NotImplementedError


class _ClockService(IClockService):
    def __init__(self, twisted_reactor):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorUNIX
        """
        IClockService.__init__(self)
        self._store = _ControlTsStore(600, int(600 * 1e6), int(5 * 1e6))

        self._twisted_reactor = twisted_reactor
        twisted_reactor.callLater(1, self._update_store, twisted_reactor)

    def _update_store(self, twisted_reactor):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorUNIX
        """
        from random import randint
        ts = long(time.time() * 1e6)
        delta = 35000
        self._store.add_entry(ts, randint(-delta, delta))
        twisted_reactor.callLater(1, self._update_store, twisted_reactor)

    def get_control_ts(self, system_ts):
        """
        :type system_ts: long
        :rtype: long | None
        """
        control_delta = self._store.lookup_nearest(system_ts)
        if control_delta is not None:
            return system_ts + control_delta

    def destroy(self):
        pass
