import ujson
import psycopg2
import psycopg2.pool
import psycopg2.extras

from edge import eventlog
from edge.com.db import DbJsonArg, DbArgs
from .types import DbServiceArgs


psycopg2.extras.register_default_json(globally=True, loads=ujson.loads)
psycopg2.extras.register_default_jsonb(globally=True, loads=ujson.loads)


class IDbService:
    def __init__(self):
        pass

    def set_log(self, detector, plate_text, log_min_ts, log_max_ts, log_jnode):
        """
        :type detector: int
        :type plate_text: str
        :type log_min_ts: int | long
        :type log_max_ts: int | long
        :type log_jnode: dict
        """
        raise NotImplementedError

    def get_log(self, detector, plate_text, plate_dist=0):
        """
        :type detector: int
        :type plate_text: str
        :type plate_dist: int
        :rtype: [dict]
        """
        raise NotImplementedError

    def del_log(self, detector, plate_text):
        """
        :type detector: int
        :type plate_text: str
        :rtype: bool
        """
        raise NotImplementedError

    def clear_log(self, detector, max_ts):
        """
        :type detector: int
        :type max_ts: long
        :rtype: bool
        """
        raise NotImplementedError

    def get_last_parking(self, plate_text, origin_uuid, min_search_time):
        """
        :type plate_text: str
        :type origin_uuid: str
        :type min_search_time: str
        """
        raise NotImplementedError


def init_service(args):
    """
    :type args: DbServiceArgs
    :rtype: IDbService
    """
    return _DbService(args.origin_id, args.db_args)


def fini_service(service):
    """
    :type service: IDbService | _DbService
    """
    service.destroy()


class _DbService(IDbService):
    def __init__(self, origin_ref, db_args):
        """
        :type origin_ref: int
        :type db_args: DbArgs
        """
        IDbService.__init__(self)
        self._origin_ref = origin_ref

        self._link = psycopg2.connect(
            host=db_args.db_host,
            port=db_args.db_port,
            user=db_args.db_user,
            password=db_args.db_pass,
            database=db_args.db_name)

    def destroy(self):
        self._link.close()

    def set_log(self, detector_ref, plate_text, log_min_ts, log_max_ts, log_jnode):
        """
        :type detector_ref: int
        :type plate_text: str
        :type log_min_ts: int | long
        :type log_max_ts: int | long
        :type log_jnode: dict | None
        :rtype: bool
        """
        # TODO: del in db, if log_jnode is None
        try:
            with self._link, self._link.cursor() as db_cursor:
                db_cursor.execute(
                    'SELECT * FROM edge.plate_motion_set(%s,%s,%s,%s,%s,%s)',
                    [self._origin_ref, detector_ref, plate_text,
                     DbJsonArg(log_jnode), log_min_ts, log_max_ts])

                row_count = db_cursor.fetchone()
                return row_count == 1

        except psycopg2.Error as x:
            eventlog.log_e('edge.plate_motion_set <%s>' % x.message)

        return False

    def get_log(self, detector_ref, plate_text, plate_dist=0):
        """
        :type detector_ref: int
        :type plate_text: str
        :type plate_dist: int
        :rtype: [(str, int, dict)] | None
        """
        try:
            with self._link, self._link.cursor() as db_cursor:
                db_cursor.execute(
                    'SELECT plate,distance,log FROM edge.plate_motion_get(%s,%s,%s,%s)',
                    [self._origin_ref, detector_ref, plate_text, plate_dist])

                return db_cursor.fetchall()

        except psycopg2.Error as x:
            eventlog.log_e('edge.plate_motion_get <%s>' % x.message)

    def del_log(self, detector_ref, plate_text):
        """
        :type detector_ref: int
        :type plate_text: str
        :rtype: bool
        """
        try:
            with self._link, self._link.cursor() as db_cursor:
                db_cursor.execute(
                    'SELECT * FROM edge.plate_motion_del(%s,%s,%s)',
                    [self._origin_ref, detector_ref, plate_text])

                row_count = db_cursor.fetchone()
                return row_count == 1

        except psycopg2.Error as x:
            eventlog.log_e('edge.plate_motion_del <%s>' % x.message)

        return False

    def clear_log(self, detector, max_ts):
        """
        :type detector: int
        :type max_ts: long
        """
        try:
            with self._link, self._link.cursor() as db_cursor:
                db_cursor.execute(
                    'SELECT * FROM edge.plate_motion_clear(%s,%s,%s)',
                    [self._origin_ref, detector, max_ts])

                row_count = db_cursor.fetchone()
                if row_count:
                    eventlog.log_d('plate_motion_clear: deleted <%s> records' % row_count)

        except psycopg2.Error as x:
            eventlog.log_e('edge.plate_motion_clear <%s>' % x.message)

    def get_last_parking(self, plate_text, origin_uuid, min_search_time):
        """
        :type plate_text: str
        :type origin_uuid: str
        :type min_search_time: str
        :rtype: bool
        """
        api_query_jnode = {
            'event': {
                'type': ['parking'],
                'origin': [origin_uuid],
                'plate-pattern': plate_text,
                'fromtime': min_search_time
            }
        }

        with self._link, self._link.cursor() as db_cursor:
            db_cursor.execute(
                'SELECT errcode,errmsg,result FROM api_event_filter(%s)',
                [ujson.dumps(api_query_jnode)])

            errcode, errmsg, api_reply_jnode = db_cursor.fetchone()

        return errcode == 0
