import os
import socket
import tarfile
import io
import functools
import ctypes

import ujson

from edge import eventlog
from edge.queue import JsonFileQueue
from edge.image.types import MemBlob, Blob, StructBlobMeta
from edge.service.about import IAboutService
from edge.service.tz import ITzService
from edge.com.bs import BsError
from edge.com.fs import silent_mkdir_r

from .stats import EventCounters
from .types import EventServiceArgs
from ..model.types import Event
from ..services.bs import IBsService
from ..services.systest import ISystestService
from ..services.sign import ISignService, SignError, BufferSizeError

from ..detectors.spot_speed.types import SpotSpeedEvidence
from ..detectors.red_light.types import RedLightEvidence
from ..detectors.parking.types import ParkingEvidence
from ..detectors.bus_lane.types import BusLaneEvidence


class IEventService:
    def __init__(self):
        pass

    def emit(self, event_args):
        """
        :type event_args: Event
        :rtype: bool
        """
        raise NotImplementedError


def init_service(twisted_reactor, counters, args, tz_service, bs_service, systest_service,
                 about_service, sign_service):
    """
    :type twisted_reactor: twisted.internet.interfaces.IReactorUNIX
    :type counters: EventCounters
    :type args: EventServiceArgs
    :type tz_service: ITzService
    :type bs_service: IBsService
    :type systest_service: ISystestService
    :type about_service: IAboutService
    :type sign_service: ISignService
    :rtype: IEventService
    """
    json_q_list = map(lambda json_q_args: JsonFileQueue(json_q_args), args.json_queue_list)
    return _EventService(twisted_reactor, counters, tz_service, bs_service, systest_service,
                         about_service, sign_service, json_q_list, args)


# noinspection PyUnusedLocal
def fini_service(service):
    """
    :type service: IEventService | _EventService
    """
    pass


def _make_fact(event_jnode, args_blob_ref, args_ua_blob_ref=None, args_ua_blob_sign=False):
    """
    :type event_jnode: dict
    :type args_blob_ref: str
    :type args_ua_blob_ref: str | None
    :type args_ua_blob_sign: bool
    :rtype: dict
    """
    def add_arch_file(name, altref=None):
        arch_file = {
            'ref': name,
            'name': name
        }

        if altref:
            arch_file['altref'] = altref

        arch_files.append(arch_file)

    def analyze_measurements():
        measurements = event_jnode['measurements']

        for measurement in measurements:
            waypoints = measurement['waypoints']
            best_images = waypoints['best'].get('images', {})
            plate_images = best_images.get('license-plates', [])

            # thumb.jpeg
            thumb_im_name = best_images.get('thumb', {}).get('name')
            if thumb_im_name is not None:
                add_arch_file(thumb_im_name, 'thumb.jpeg')

            # scene.jpeg
            scene_im_name = best_images.get('scene', {}).get('name')
            if scene_im_name is not None:
                add_arch_file(scene_im_name, 'scene.jpeg')

            # vehicle.jpeg
            vehicle_im_name = best_images.get('vehicle', {}).get('name')
            if vehicle_im_name is not None:
                add_arch_file(vehicle_im_name, 'vehicle.jpeg')

            # overview.jpeg
            overview_im_name = best_images.get('overview', {}).get('name', None)
            if overview_im_name is not None:
                add_arch_file(overview_im_name, 'overview.jpeg')

            # plate.jpeg
            map(lambda plate_im: add_arch_file(plate_im['name'], 'plate.jpeg'),
                plate_images)

    def analyze_evidence():
        e_jnode = event_jnode.get('evidence', {})
        em_jnode = e_jnode.get('measurements')
        if not em_jnode:
            return

        emw_jnodes = em_jnode.get('waypoints', {})
        for waypoint in emw_jnodes.itervalues():
            if type(waypoint) != list:
                waypoint = [waypoint]

            for w_jnode in waypoint:
                wi_jnodes = w_jnode.get('images', {})
                for wi_jnode in wi_jnodes.itervalues():
                    add_arch_file(wi_jnode['name'])

        emv_jnodes = em_jnode.get('videos', [])
        for emv_jnode in emv_jnodes:
            add_arch_file(emv_jnode['name'])

    arch_files = []
    add_arch_file(args_blob_ref, 'args.json')

    if args_ua_blob_ref is not None:
        add_arch_file(args_ua_blob_ref, 'args-ua.json')

        if args_ua_blob_sign:
            sign_name = os.path.splitext(args_ua_blob_ref)[0] + '.sign'
            cert_name = os.path.splitext(args_ua_blob_ref)[0] + '.crt'
            add_arch_file(sign_name, 'args-ua.sign')
            add_arch_file(cert_name, 'args-ua.crt')

    analyze_measurements()
    analyze_evidence()

    jnode = {
        'args': event_jnode,
        'dist': arch_files
    }

    return jnode


def _pack_args_ua(event_args, tz_service, about_service):
    """
    :type event_args: Event
    :type tz_service: ITzService
    :type about_service: IAboutService
    :rtype: dict
    """
    base_url = about_service.base_url
    if base_url:
        base_url = '{}/{}/{}/'.format(base_url, event_args.uuid, 'by-ref')

    base = event_args.base
    bw = base.measurements[0].best_waypoint
    ev = event_args.evidence

    mmr_to_ua_class_map = {
        'CAR': '1',
        'HT': '21',
        'LT': '2',
        'VAN': '11',
        'MINIVAN': '12',
        'CROSSOVER': '13',
        'OFFROAD': '14',
        'BUS': '3',
        'UNKNOWN': '0'
    }

    def ts_to_datetime(ts):
        """
        :type ts: int
        :rtype: str
        """
        return tz_service.get_datetime_isostr(ts / 1e6)

    direction = bw.vehicle.direction
    jnode = {
        "uuid": event_args.uuid,
        "datetime":	ts_to_datetime(ev.ts),
        "type": ev.type,
        "origin": {
            "uuid": ev.origin.uuid,
            "name": about_service.get_hw_model(),
            "serial": about_service.get_hw_serial(),
            "address": ev.origin.location.place,
            "type": {
                "t_install": 1,  # stationary
                "t_measure": [1, 3]  # radar + camera
            }
        },
        "vehicle": {
            "plate": {
                "country":	bw.vehicle.plate_county_numeric,
                "text": bw.vehicle.plate_text
            },
            "params":
            [
                {
                    "key":	"direction",
                    "value": str(direction)
                },
                {
                    "key":	'color',
                    "value": '000'
                },
                {
                    "key":	"type",
                    "value": mmr_to_ua_class_map.get(bw.vehicle.vehicle_class, '0')
                },
                {
                    "key":	"mark",
                    "value": bw.vehicle.vehicle_make or 'unknown'
                },
                {
                    "key":	"model",
                    "value": bw.vehicle.vehicle_model or 'unknown'
                }
            ]
        },
        "limits": [],
        "measurements":	[],
        "media": []
    }

    def append_measurement(key, value):
        """
        :type key: str
        :type value: str
        """
        jnode['measurements'].append({'key': key, 'value': value})

    def append_limits(key, value):
        """
        :type key: str
        :type value: str
        """
        jnode['limits'].append({'key': key, 'value': value})

    def append_blob(blob, blob_type):
        """
        :type blob: MemBlob
        :type blob_type: str
        """
        jnode['media'].append({'type': blob_type, 'baseurl': base_url, 'name': blob.name,
                               'hash': blob.hash, 'datetime': ts_to_datetime(blob.min_ts)})

    # add geolocation if exist
    if ev.origin.geolocation:
        jnode['origin']['location'] = ev.origin.geolocation.serialize_dict()

    # add limits
    if ev.type == ev.NAME_SPOT_SPEED:
        ss_ev = ev  # type:SpotSpeedEvidence
        append_limits('speed-limit', str(ss_ev.rule_args.limits.speed_limit))
        append_limits('max-overspeed', str(ss_ev.rule_args.limits.over_speed))

    elif ev.type == ev.NAME_RED_LIGHT:
        # rl_ev = ev  # type: RedLightEvidence
        append_limits('min-red-time', '0')

    elif ev.type == ev.NAME_PARKING:
        pk_ev = ev  # type: ParkingEvidence
        append_limits('max-standing-time', str(pk_ev.rule_args.limits.max_standing_time))

    # add measurements
    if ev.type == ev.NAME_SPOT_SPEED:
        ss_ev = ev  # type:SpotSpeedEvidence
        append_measurement('speed-main', str(ss_ev.speed))
        append_measurement('speed-control', str(ss_ev.cam_speed))

    elif ev.type == ev.NAME_RED_LIGHT:
        rl_ev = ev  # type: RedLightEvidence
        green_amder_t, red_t, red_amber_t = rl_ev.get_ukr_json_ev()
        append_measurement('green-amber-time', str(green_amder_t))
        append_measurement('red-amber-time', str(red_amber_t))
        append_measurement('red-time', str(red_t))

    elif ev.type == ev.NAME_PARKING:
        pk_ev = ev  # type: ParkingEvidence
        append_measurement('standing-time', str(pk_ev.standing_time))

    # add media
    if ev.type == ev.NAME_SPOT_SPEED:
        ss_ev = ev  # type:SpotSpeedEvidence
        best_blobs = ss_ev.best_waypoint.get_blobs()
    elif ev.type == ev.NAME_BUS_LANE:
        bl_ev = ev  # type:BusLaneEvidence
        best_blobs = bl_ev.best_waypoint.get_blobs()
        inside_blobs = bl_ev.waypoints[1].get_blobs()
        inside_scene = filter(lambda blob: blob.profile == 'scene', inside_blobs)[0]
        append_blob(inside_scene, 'inside')

    else:
        best_blobs = bw.get_blobs()

    pl_blob = filter(lambda blob: blob.profile == 'plate', best_blobs)[0]
    th_blob = filter(lambda blob: blob.profile == 'thumb', best_blobs)[0]
    vh_blob = filter(lambda blob: blob.profile == 'vehicle', best_blobs)[0]

    append_blob(pl_blob, 'plate')
    append_blob(th_blob, 'thumb')
    append_blob(vh_blob, 'body')

    ev_wps = ev.waypoints
    near_wp = None
    far_wp = None
    if len(ev_wps) > 1:
        # TODO: find wp with min\max ts instead using index in case of non sorted lists

        if direction.dir == direction.RECEDING:
            if ev_wps[0].ts < ev_wps[-1].ts:
                near_wp = ev_wps[0]
                far_wp = ev_wps[-1]
            else:
                near_wp = ev_wps[-1]
                far_wp = ev_wps[0]

        elif direction.dir == direction.APPROACHING:
            if ev_wps[0].ts < ev_wps[-1].ts:
                near_wp = ev_wps[-1]
                far_wp = ev_wps[0]
            else:
                near_wp = ev_wps[0]
                far_wp = ev_wps[-1]

    if None not in [near_wp, far_wp]:
        near_full_blob = filter(lambda blob: blob.profile == 'scene', near_wp.get_blobs())[0]
        far_full_blob = filter(lambda blob: blob.profile == 'scene', far_wp.get_blobs())[0]
        append_blob(near_full_blob, 'close')
        append_blob(far_full_blob, 'far')

    video_blobs = ev.videos
    if len(video_blobs):
        append_blob(video_blobs[0], 'video')

    return jnode


class _EventService(IEventService):
    def __init__(self, twisted_reactor, counters, tz_service, bs_service, systest_service,
                 about_service, sign_service, json_q_list, args):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorUNIX
        :type counters: EventCounters
        :type tz_service: ITzService
        :type bs_service: IBsService
        :type systest_service: ISystestService
        :type about_service: IAboutService
        :type sign_service: ISignService
        :type json_q_list: list[JsonFileQueue]
        :type args: EventServiceArgs
        """
        IEventService.__init__(self)
        self._reactor = twisted_reactor
        self._counters = counters

        self._tz_service = tz_service
        self._bs_service = bs_service
        self._systest_service = systest_service
        self._about_service = about_service
        self._sign_service = sign_service
        self._json_q_list = json_q_list
        self._devices_jnode = self._get_devices(about_service)

        self._args_ua_required = args.ua_json_required
        self._local_arch_enabled = args.local_arch_enabled
        self._local_arch_base_dir = args.local_arch_base_dir

        self._hash_buffer_size = 100

    @staticmethod
    def _get_devices(about_service):
        """
        :rtype: list[dict]
        """
        device = {'hostname': socket.gethostname()}
        hw = about_service.hw_info
        if any(hw):
            device['hw'] = hw

        sw = about_service.sw_info
        if sw:
            device['sw'] = sw

        sw_name = about_service.sw_name
        if sw_name:
            device['sw-name'] = sw_name

        return [device]

    def _hash_blob(self, blob):
        """
        :type blob: Blob
        :rtype: None
        """
        hash_buffer = ctypes.create_string_buffer('', size=self._hash_buffer_size)
        hash_size = 0
        try:
            hash_size = self._sign_service.hash_blob(blob.file_data, hash_buffer)
        except BufferSizeError as x:
            self._hash_buffer_size = x.min_size
            hash_buffer = ctypes.create_string_buffer('', size=self._hash_buffer_size)
            hash_size = self._sign_service.hash_blob(blob.file_data, hash_buffer)

        except SignError:
            eventlog.log_e('fail: hash <profile={}>'.format(blob.profile))

        if hash_size > 0:
            blob.hash = self._sign_service.hash_type, hash_buffer.raw[:hash_size]

    def _emit2(self, event_args, blobs_to_test, args_blob, args_ua_blob, media_blobs, args_jnode, args_ua_signed):
        def get_tar_path(base_dir, ts):
            """
            :type base_dir: str
            :type ts: long
            :rtype: str
            """
            d = self._tz_service.get_datetime(ts/1e6)
            path = os.path.join(base_dir, str(d.year), str(d.month).zfill(2),
                                str(d.day).zfill(2), str(d.hour).zfill(2),
                                str(d.minute).zfill(2))
            return path

        def export_event_to_gzip():
            ts = event_args.ts
            filedir = get_tar_path(self._local_arch_base_dir, ts)
            silent_mkdir_r(filedir)
            args_blob_id = args_blob.name.split('_')[1].split(".")[0]
            tar_name = os.path.join(filedir, '{}.tar'.format(args_blob_id))

            with tarfile.open(name=tar_name, mode='w') as tar:
                args_info = tarfile.TarInfo('args.json')
                args_info.size = len(args_blob.file_data)
                tar.addfile(args_info, io.BytesIO(args_blob.file_data))

                if args_ua_blob is not None:
                    args_info = tarfile.TarInfo('args-ua.json')
                    args_info.size = len(args_ua_blob.file_data)
                    tar.addfile(args_info, io.BytesIO(args_ua_blob.file_data))

                for media_blob in media_blobs:
                    if media_blob.file_path is None:
                        tar_info = tarfile.TarInfo(media_blob.name)
                        tar_info.size = len(media_blob.file_data)
                        tar.addfile(tar_info, io.BytesIO(media_blob.file_data))
                    else:
                        tar.add(media_blob.file_path, media_blob.name)

        try:

            if args_ua_blob:
                self._bs_service.save(event_args.sender, args_ua_blob)
                self._hash_blob(args_ua_blob)
                blobs_to_test.append(args_ua_blob)

                fact_jnode = _make_fact(args_jnode, args_blob.name, args_ua_blob.name, args_ua_signed)
            else:
                fact_jnode = _make_fact(args_jnode, args_blob.name)

            skip_event = False
            for blob in blobs_to_test:
                blob_id = blob.name.split('_')[1].split('.')[0]
                blob_size = self._bs_service.get_size(event_args.sender, blob_id, blob.profile)
                blob_size2 = os.stat(blob.file_path).st_size if blob.file_path else len(blob.file_data)
                if blob_size != blob_size2:
                    skip_event = True
                    eventlog.log_i('blob_id {} / {} profile {:12s} blob_size {:7d} / {:7d} ALARM'.format(
                        blob_id, hex(int(blob_id)), blob.profile, blob_size, blob_size2))

            if not skip_event:
                if self._local_arch_enabled:
                    export_event_to_gzip()

                map(lambda json_q: json_q.add_jnode(fact_jnode, args_blob.name), self._json_q_list)
                self._counters.inc_fact()
            else:
                eventlog.log_w('fail: skip event due to ALARM-s')

        except BsError:
            eventlog.log_w('fail: save-blob (skip event)')

        except Exception:
            eventlog.log_x()

    def emit(self, event_args):
        """
        :type event_args: Event
        :rtype: bool
        """
        def handle_sign(event_args2, blobs_to_test2, args_blob2, args_ua_blob2, media_blobs2, args_jnode2,
                        blob_data, sign_data):
            """
            :type event_args2:
            :type blobs_to_test2:
            :type args_blob2:
            :type args_ua_blob2:
            :type media_blobs2:
            :type args_jnode2:
            :type blob_data: str | None
            :type sign_data: str | None
            :return:
            """
            args_ua_signed = sign_data is not None

            if args_ua_signed:
                if blob_data != args_ua_blob2.file_data:
                    eventlog.log_w("fail: blob_data != args_ua_blob.file_data")

                crt = self._sign_service.cert
                if crt:
                    args_ua_blob2.meta = StructBlobMeta(sign_data, len(sign_data), crt)
                else:
                    args_ua_blob2.meta = StructBlobMeta(sign_data, len(sign_data))

            self._emit2(event_args2, blobs_to_test2, args_blob2, args_ua_blob2, media_blobs2,
                        args_jnode2, args_ua_signed)

        def sign_blob(blob_to_sign, callback):
            """
            :type blob_to_sign: Blob
            :type callback: (str | None, str | None) -> None

            :rtype: None
            """
            try:
                return self._sign_service.sign_blob(blob_to_sign.file_data, callback)

            except SignError:
                pass

            except Exception:
                eventlog.log_x()

            callback(None, None)

        # noinspection PyBroadException
        try:
            # save media
            media_blobs = event_args.get_blobs()
            map(lambda m_blob: self._bs_service.save(event_args.sender, m_blob), media_blobs)
            map(self._hash_blob, media_blobs)

            # save args.json
            args_jnode = event_args.serialize(self._tz_service)
            self._systest_service.serialize(args_jnode)
            args_jnode['origin']['devices'] = self._devices_jnode
            args_blob = MemBlob('args', event_args.ts, ujson.dumps(args_jnode), file_ext='json')

            self._bs_service.save(event_args.sender, args_blob)
            self._hash_blob(args_blob)

            # save args-ua.json
            blobs_to_test = media_blobs + [args_blob]
            args_ua_blob = None
            if self._args_ua_required:
                args_ua_jnode = _pack_args_ua(event_args, self._tz_service, self._about_service)
                args_ua_blob = MemBlob('args', event_args.ts, ujson.dumps(args_ua_jnode), file_ext='json')

                if event_args.type in [Event.TYPE_BUS_LANE,
                                       Event.TYPE_SPOT_SPEED,
                                       Event.TYPE_SECTION_SPEED,
                                       Event.TYPE_WRONG_WAY,
                                       Event.TYPE_RED_LIGHT,
                                       Event.TYPE_PARKING,
                                       Event.TYPE_WANTED]:

                    handler = functools.partial(
                        handle_sign, event_args, blobs_to_test,
                        args_blob, args_ua_blob, media_blobs, args_jnode)

                    sign_blob(args_ua_blob, handler)
                else:
                    self._emit2(event_args, blobs_to_test, args_blob, args_ua_blob, media_blobs, args_jnode, False)

            else:
                self._emit2(event_args, blobs_to_test, args_blob, args_ua_blob, media_blobs, args_jnode, False)

        except BsError:
            eventlog.log_w('fail: save-blob (skip event)')
            return False

        except Exception:
            eventlog.log_x()
            return False

        return True
