from edge import eventlog
from edge.types import JnodeError

from .jnode.types import JsonLeddarServiceArgs
from edge.model.types import TimeSpan


class ILeddarService:
    def __init__(self):
        pass

    def update(self, react_jnode):
        pass

    def ref(self):
        """
        :rtype: str | None
        """
        return None

    def check(self, min_ts, max_ts):
        """
        :type min_ts: long
        :type max_ts: long
        :rtype: list[TimeSpan]
        """
        raise NotImplementedError

    def destroy(self):
        pass


def init_service(args):
    """
    :type args: JsonLeddarServiceArgs | None
    :rtype: ILeddarService
    """
    if args is None:
        rl = _StubService()
    else:
        rl = _LeddarService(args)
    return rl


def fini_service(service):
    """
    :type service: ILeddarService | _LeddarService
    """
    service.destroy()


class _StubService(ILeddarService):
    def __init__(self):
        ILeddarService.__init__(self)

    def check(self, min_ts, max_ts):
        return []


class LeddarLog:
    def __init__(self, time, active):
        """
        :type time: TimeSpan
        :type active: list[TimeSpan]
        """
        self.active = active
        self.time = time

    @staticmethod
    def deserialize(jnode, leddar_ref):
        """
        :type jnode: dict
        :type leddar_ref: str
        :rtype: JsonTlArgs

        :raise JnodeError:
        """
        ts_jnode = jnode.get('ts')

        if type(ts_jnode) != list or len(ts_jnode) != 2:
            eventlog.log_e_param('#/ts')
            raise JnodeError

        time_interval = TimeSpan(*ts_jnode)

        leddar_list_jnode = jnode.get('leddar')
        if type(leddar_list_jnode) != list:
            eventlog.log_e_param('#/leddar')
            raise JnodeError

        for leddar_jnode in leddar_list_jnode:
            if leddar_jnode['ref'] != leddar_ref:
                continue

            active_jnode = leddar_jnode['active']
            active_intervals = map(TimeSpan.create, active_jnode)
            return LeddarLog(time_interval, active_intervals)

        raise JnodeError

    def check(self, time_span):
        """
        :type time_span: TimeSpan
        :rtype: list[TimeSpan]
        """
        spans = filter(lambda x: x.check_overlap(time_span), self.active)
        return spans


class _LeddarService(ILeddarService):
    """
    :type _leddar_log: LeddarLog | None
    """
    def __init__(self, args):
        """
        :type args: JsonLeddarServiceArgs
        """
        ILeddarService.__init__(self)

        self._leddar_log = None
        self._leddar_ref = args.ref

    def ref(self):
        return self._leddar_ref

    def update(self, react_jnode):
        try:
            self._leddar_log = LeddarLog.deserialize(react_jnode, self._leddar_ref)
        except Exception as x:
            eventlog.log_x()
            return

    def check(self, min_ts, max_ts):
        """
        :type min_ts: long
        :type max_ts: long
        :rtype: list[TimeSpan]
        """
        if self._leddar_log is None:
            return []

        time_interval = TimeSpan(min_ts, max_ts)
        if not self._leddar_log.time.check_overlap(time_interval):
            # there is no stats for [cts ,lts]
            return []

        violation_list = self._leddar_log.check(time_interval)
        return violation_list
