import ctypes
import os

from edge import eventlog
from edge.com import fs
from edge.image.image import ImRgba, ImBuffer
from edge.image.types import FileBlob, JsonBlobMeta
from glob import glob
from tempfile import NamedTemporaryFile
from .types import VsServiceArgs


_IMAGE_POOL_MAX = 1
_VIDEO_POOL_MAX = 3


class VsImageInfo:
    def __init__(self, ts):
        """
        :type ts: long
        """
        self._ts = ts

    @property
    def ts(self):
        """
        :return: value in usec
        :rtype: long
        """
        return self._ts


class VsError(RuntimeError):
    """
    Generic Error.
    """
    pass


class VsImageBufferError(VsError):
    def __init__(self, min_w, min_h):
        """
        :type min_w: int
        :type min_h: int
        """
        self.min_w = min_w
        self.min_h = min_h


class VsImageBufferError2(VsError):
    def __init__(self, size):
        """
        :type size: long
        """
        self.min_size = size


class VsSessionAbsentError(VsError):
    """
    Session is absent in the pool.
    """
    pass


class VsSessionExpiredError(VsError):
    """
    Session is expired.
    Session should be closed and opened again.
    """
    pass


class IVsService:
    def __init__(self):
        pass

    def destroy(self):
        raise NotImplementedError

    def export_image_rgba(self, im_rgba, im_infos, bs_chan=None):
        """
        :type im_rgba: ImRgba
        :type im_infos: list[VsImageInfo]
        :type bs_chan: str | None

        :rtype: VsImageInfo | None

        :raise VsImageBufferError:
        :raise VsError:
        """
        raise NotImplementedError

    def export_image_jpeg(self, im_rgba, im_infos, bs_chan=None):
        """
        :type im_rgba: ImBuffer
        :type im_infos: list[VsImageInfo]
        :type bs_chan: str | None

        :rtype: VsImageInfo | None

        :raise VsImageBufferError2:
        :raise VsError:
        """
        raise NotImplementedError

    def export_image_rgba2(self, im_rgba, im_infos):
        """
        :type im_rgba: ImRgba
        :type im_infos: list[VsImageInfo]

        :rtype: VsImageInfo | None

        :raise VsError:
        """
        while True:
            try:
                return self.export_image_rgba(im_rgba, im_infos)
            except VsImageBufferError as x:
                min_w, min_h = x.min_w, x.min_h

            im_rgba.reset(im_rgba.ts, min_w, min_h, im_rgba.dh)

    def export_video(self, min_ts, max_ts, max_w, max_h, rec_quality, handler):
        """
        :param min_ts: ms since 01.01.1970 00:00 UTC
        :type min_ts: long

        :param max_ts: ms since 01.01.1970 00:00 UTC
        :type max_ts: long

        :param max_w: valid values in range [16; 32752], multiple 16.
        :type max_w: int

        :param max_h: valid values in range [16; 32752], multiple 16.
        :type max_h: int

        :param rec_quality: Constant Bit Rate (CBR), Kbit/s. Valid values in range [100; 102400].
        :type rec_quality: int

        :type handler: (FileBlob | None) -> None

        :rtype: None
        :raise VsSessionAbsentError:
        :raise VsError:
        """
        raise NotImplementedError


class _Api:
    RC_DONE = 0
    RC_FAIL = -1
    RC_DENY = -2
    RC_CONN = -3

    LOG_N = 0
    LOG_E = 1
    LOG_W = 2
    LOG_I = 3
    LOG_D = 4
    LOG_T = 5

    IMAGE_GRAY = 1
    IMAGE_RGBA = 2
    IMAGE_JPEG = 3

    VIDEO_H264_AVI = 1
    VIDEO_H264_MKV = 2
    VIDEO_H264_MP4 = 3

    _lib = ctypes.CDLL(glob('/opt/edge/lib/libblobstore*.so')[0])

    # int
    # BsvOpen(uint16_t aStore, uint8_t aLogLevel, void **aSession);

    open = _lib.BsvOpen
    open.argtypes = (ctypes.c_uint16, ctypes.c_uint8, ctypes.POINTER(ctypes.c_voidp))
    open.restype = ctypes.c_int

    # int
    # BsvLoadImage(void *aSession, const char *aChannel, uint64_t anImTs, uint8_t anImFormat, uint8_t *anImData,
    #     uint32_t *anImSize,uint16_t *anImWidth, uint16_t *anImHeight);

    load = _lib.BsvLoadImage
    load.argtypes = (
        ctypes.c_voidp, ctypes.c_char_p, ctypes.c_uint64, ctypes.c_uint8, ctypes.POINTER(ctypes.c_uint8),
        ctypes.POINTER(ctypes.c_uint32), ctypes.POINTER(ctypes.c_uint16), ctypes.POINTER(ctypes.c_uint16))

    load.restype = ctypes.c_int

    # int
    # BsvLoadImageCoded(void *aSession, const char *aChannel, uint64_t *anImTs, uint8_t anImFormat, uint8_t *anImData,
    #     uint32_t *anImSize);

    loadCoded = _lib.BsvLoadImageCoded
    loadCoded.argtypes = (
        ctypes.c_voidp, ctypes.c_char_p, ctypes.POINTER(ctypes.c_uint64), ctypes.c_uint8, ctypes.POINTER(ctypes.c_uint8),
        ctypes.POINTER(ctypes.c_uint32))

    loadCoded.restype = ctypes.c_int

    # void
    # typedef void (*bsv_load_record_handler_fn)(int aStat, uint64_t aRecMinTs,
    #               uint64_t aRecMaxTs, const char *aRecPath,
    #               void *aHandlerArg);

    LoadRecordHandler = ctypes.CFUNCTYPE(
        None, ctypes.c_int, ctypes.c_uint64,
        ctypes.c_uint64, ctypes.c_char_p,
        ctypes.c_voidp)

    # int
    # BsvLoadRecordAsync(void *aSession, const char *aChannel,
    #     uint64_t aRecMinTs, uint64_t aRecMaxTs,
    #     uint8_t aRecFormat, uint32_t aRecQuality, uint16_t aRecSkipRate,
    #     uint16_t aRecMaxWidth, uint16_t aRecMaxHeight, const char *aRecPath,
    #     bsv_load_record_handler_fn aHandlerFn, void *aHandlerArg);

    save = _lib.BsvLoadRecordAsync
    save.argtypes = (
        ctypes.c_voidp, ctypes.c_char_p,
        ctypes.c_uint64, ctypes.c_uint64,
        ctypes.c_uint8, ctypes.c_uint32, ctypes.c_uint16,
        ctypes.c_uint16, ctypes.c_uint16, ctypes.c_char_p,
        LoadRecordHandler, ctypes.c_voidp)

    save.restype = ctypes.c_int

    # void
    # BsvClose(void *aSession);
    close = _lib.BsvClose
    close.argtype = ctypes.c_voidp

    def __init__(self):
        pass


class _VsSession:
    def __init__(self, handle, chan):
        """
        :type handle: ctypes.c_void_p | None
        :type chan: str
        """
        self._chan = chan
        self._handle = handle
        self.closed = handle is None

    @staticmethod
    def open(store, chan, log_level=_Api.LOG_I):
        """
        :type store: int
        :type chan: str
        :type log_level: int

        :rtype: _VsSession
        :raise VsSessionExpiredError:
        """
        handle = ctypes.c_void_p(0)
        rc = _Api.open(store, log_level, handle)

        if rc != _Api.RC_DONE:
            eventlog.log_e('fail: BsvOpen (rc:%s)' % rc)
            raise VsSessionExpiredError

        return _VsSession(handle=handle, chan=chan)

    def export_image_rgba(self, im_ts, im_buffer, bs_chan=None):
        """
        :param im_ts: value in msec
        :type im_ts: long
        :type bs_chan: str | None

        :type im_buffer: str

        :rtype: (int, int)
        :raise VsSessionExpiredError:
        :raise VsError:
        """
        if self.closed:
            eventlog.log_e('fail: session closed')
            raise VsSessionExpiredError

        im_data = ctypes.cast(im_buffer, ctypes.POINTER(ctypes.c_uint8))
        im_size = ctypes.c_uint32(ctypes.sizeof(im_buffer))
        im_w, im_h = ctypes.c_uint16(), ctypes.c_uint16()

        chan = self._chan if bs_chan is None else bs_chan
        rc = _Api.load(
            self._handle, chan, ctypes.c_uint64(im_ts), ctypes.c_uint8(_Api.IMAGE_RGBA),
            im_data, im_size, im_w, im_h)

        if im_size.value > ctypes.sizeof(im_buffer):
            eventlog.log_e('fail: buffer too small')
            return im_w.value, im_h.value

        if rc == _Api.RC_CONN:
            self.close()
            eventlog.log_e('fail: BsvLoadImage (kBS_CONN)')
            raise VsSessionExpiredError

        elif rc != _Api.RC_DONE:
            eventlog.log_e('fail: BsvLoadImage (rc:%s)' % rc)
            raise VsError

        return im_w.value, im_h.value

    def export_image_jpeg(self, im_ts, im_buffer, bs_chan=None):
        """
        :param im_ts: value in msec
        :type im_ts: long
        :type bs_chan: str | None

        :type im_buffer: str

        :rtype: (int, long)
        :raise VsSessionExpiredError:
        :raise VsError:
        """
        if self.closed:
            eventlog.log_e('fail: session closed')
            raise VsSessionExpiredError

        im_data = ctypes.cast(im_buffer, ctypes.POINTER(ctypes.c_uint8))
        im_size = ctypes.c_uint32(ctypes.sizeof(im_buffer))
        im_ts_c = ctypes.c_uint64(im_ts)

        chan = self._chan if bs_chan is None else bs_chan
        rc = _Api.loadCoded(
            self._handle, chan, im_ts_c, ctypes.c_uint8(_Api.IMAGE_JPEG),
            im_data, im_size)

        if im_size.value > ctypes.sizeof(im_buffer):
            eventlog.log_e('fail: buffer too small %s' % im_size.value)
            return im_size.value, im_ts_c.value

        if rc == _Api.RC_CONN:
            self.close()
            eventlog.log_e('fail: BsvLoadImage (kBS_CONN)')
            raise VsSessionExpiredError

        elif rc != _Api.RC_DONE:
            eventlog.log_e('fail: BsvLoadImage (rc:%s) %s' % (rc, im_size.value))
            raise VsError

        return im_size.value, im_ts_c.value

    def export_video(self, min_ts, max_ts, max_w, max_h, rec_q, file_path, cb, cb_arg):
        """
        :param min_ts: ms since 01.01.1970 00:00 UTC
        :type min_ts: long

        :param max_ts: ms since 01.01.1970 00:00 UTC
        :type max_ts: long

        :param max_w: valid values in range [16; 32752], multiple 16.
        :type max_w: int

        :param max_h: valid values in range [16; 32752], multiple 16.
        :type max_h: int

        :param rec_q: Constant Bit Rate (CBR), Kbit/s. Valid values in range [100; 102400].
        :type rec_q: int

        :type file_path: str

        :param cb: _Api.LoadRecordHandler
        :type cb: Any
        :type cb_arg: int

        :rtype: None
        :raise VsSessionExpiredError:
        :raise VsError:
        """
        if self.closed:
            eventlog.log_e('fail: session closed')
            raise VsSessionExpiredError

        rc = _Api.save(
            self._handle, self._chan,
            ctypes.c_uint64(min_ts),
            ctypes.c_uint64(max_ts),
            ctypes.c_uint8(_Api.VIDEO_H264_MP4),
            ctypes.c_uint32(rec_q),
            ctypes.c_uint16(1),  # skip-rate
            ctypes.c_uint16(max_w),
            ctypes.c_uint16(max_h),
            ctypes.cast(file_path, ctypes.POINTER(ctypes.c_char)),
            cb, ctypes.c_void_p(cb_arg))

        if rc == _Api.RC_CONN:
            self.close()
            eventlog.log_e('fail: BsvLoadRecordAsync (kBS_CONN)')
            raise VsSessionExpiredError

        elif rc != _Api.RC_DONE:
            eventlog.log_e('fail: BsvLoadRecordAsync (rc:%s)' % rc)
            raise VsError

    def close(self):
        if not self.closed:
            _Api.close(self._handle)
            self.closed = True


class _VsSessionPool:
    _LOG_NAME_TO_CODE_MAP = {
        'none': _Api.LOG_N,
        'error': _Api.LOG_E,
        'warn': _Api.LOG_W,
        'info': _Api.LOG_I,
        'debug': _Api.LOG_D,
        'trace': _Api.LOG_T
    }

    def __init__(self, bs_store, bs_chan, bs_log, max_count):
        """
        :type bs_store: int
        :type bs_chan: str
        :type bs_log: str
        """
        log_code = _VsSessionPool._LOG_NAME_TO_CODE_MAP.get(bs_log, 'info')
        self._conf = bs_store, bs_chan, log_code
        self._sessions = map(
            lambda _: _VsSession(handle=None, chan=bs_chan),
            range(max_count))

    def __del__(self):
        """
        :rtype: None
        """
        map(lambda x: x.close(), self._sessions)

    def get_session(self):
        """
        :rtype: _VsSession
        :raise VsSessionAbsentError:
        :raise VsError:
        """
        if not self._sessions:
            eventlog.log_e('fail: session absent')
            raise VsSessionAbsentError

        session = self._sessions.pop()
        try:
            attempt = 0
            while attempt < 3:
                attempt += 1
                try:
                    if session.closed:
                        session = _VsSession.open(*self._conf)

                    eventlog.log_d('done: session:%s' % id(session))
                    return session

                except VsSessionExpiredError:
                    session.close()

                # todo: consider to sleep a little bit
                eventlog.log_e('fail: open session (attempt:%s/3)' % attempt)

            raise VsError

        except VsError:
            self._sessions.append(session)
            raise

        except Exception:
            self._sessions.append(session)
            raise

    def put_session(self, session):
        """
        :type session: _VsSession
        :rtype: None
        """
        assert session not in self._sessions
        self._sessions.append(session)
        eventlog.log_d('done: session:%s' % id(session))


class _VsImageSessionPool(_VsSessionPool):
    pass


class _VsVideoSessionPool(_VsSessionPool):
    pass


class _VsVideoWork:
    """
    :type _session: _VsSession | None
    """
    _ref_to_self_map = dict()

    def __init__(self, twisted_reactor, tmp_path, mk_session_fn, rm_session_fn,
                 min_ts, max_ts, max_w, max_h, req_q, handler):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorUNIX
        :type tmp_path:str
        :type mk_session_fn: () -> _VsSession
        :type rm_session_fn: (_VsSession) -> None

        :type min_ts: long
        :type max_ts: long
        :type max_w: int
        :type max_h: int
        :type req_q: int
        :type handler: (FileBlob | None) -> None
        """
        self._ref = id(self)

        self._reactor = twisted_reactor
        self._handler = handler
        self._mk_session_fn = mk_session_fn
        self._rm_session_fn = rm_session_fn
        self._session = None

        #
        self._tmp_file = NamedTemporaryFile(dir=tmp_path, prefix='bsVid')
        eventlog.log_d('done: mk tmpfile (ref:{}, name:{})'.format(self._ref, self._tmp_file.name))

        #
        def c_call_handler_py(rec_stat, rec_min_ts, rec_max_ts, rec_path, work_ref):
            """
            The handler is called from video-provider's thread. Do synchronize with the twisted reactor

            :type rec_stat: int
            :type rec_min_ts: long
            :type rec_max_ts: long
            :type rec_path: str
            :type work_ref: int
            :rtype: None
            """
            self._reactor.callFromThread(_VsVideoWork._twisted_handler,
                                         rec_stat, rec_min_ts, rec_max_ts, rec_path, work_ref)

        self._c_call_handler = _Api.LoadRecordHandler(c_call_handler_py)
        self._c_call_args = (min_ts, max_ts, max_w, max_h, req_q, self._tmp_file.name,
                             self._c_call_handler, self._ref)

    def __del__(self):
        self._tmp_file = None
        if _VsVideoWork._ref_to_self_map.pop(self._ref, None):
            eventlog.log_w('done: fini video-work (ref:%s)' % self._ref)

    @staticmethod
    def _twisted_handler(rec_stat, rec_min_ts, rec_max_ts, rec_path, work_ref):
        """
        :type rec_stat: int
        :type rec_min_ts: long
        :type rec_max_ts: long
        :type rec_path: str
        :type work_ref: int
        :rtype: None
        """
        eventlog.log_d('exec: fini video-work (ref:%s)' % work_ref)
        self = _VsVideoWork._ref_to_self_map.pop(work_ref, None)  # type: _VsVideoWork
        if self is None:
            eventlog.log_w('fail: fini video-work (ref:%s)' % work_ref)
            return

        else:
            eventlog.log_d('done: fini video-work (ref:%s)' % work_ref)

        #
        if rec_stat == _Api.RC_CONN:
            eventlog.log_e('fail: BsvLoadRecordAsync (kBS_CONN)')
            self._session.close()
            video_blob = None

        elif rec_stat != _Api.RC_DONE:
            eventlog.log_e('fail: BsvLoadRecordAsync (stat:%s)' % rec_stat)
            video_blob = None

        else:
            eventlog.log_d('done: BsvLoadRecordAsync (stat:%s)' % rec_stat)
            max_w = self._c_call_args[2]
            max_h = self._c_call_args[3]
            video_blob_meta = JsonBlobMeta({'size': [max_w, max_h, os.path.getsize(rec_path)]})
            video_blob = FileBlob(rec_min_ts * 1000, rec_max_ts * 1000, self._tmp_file, meta=video_blob_meta)

        self._rm_session_fn(self._session)
        self._session = None

        # noinspection PyBroadException
        try:
            self._c_call_handler = None
            self._c_call_args = None
            self._tmp_file = None
            self._handler(video_blob)
            eventlog.log_d('done: ref:{}, blob:{}'.format(
                self._ref, video_blob.file_path if video_blob else 'none'))
        except Exception:
            eventlog.log_x()

    def run_async(self):
        """
        :rtype: None
        :raise VsSessionAbsentError:
        :raise VsError:
        """
        self._session = self._mk_session_fn()

        _VsVideoWork._ref_to_self_map[self._ref] = self
        eventlog.log_d('exec: init video-work (ref:%s)' % self._ref)

        # noinspection PyBroadException
        try:
            self._session.export_video(*self._c_call_args)
            eventlog.log_d('done: init video-work (ref:%s)' % self._ref)
            return
        except VsError:
            eventlog.log_e('fail: init video-work (ref:%s, session error' % self._ref)
        except Exception:
            eventlog.log_x()

        self._c_call_handler = None
        self._c_call_args = None
        self._tmp_file = None
        del _VsVideoWork._ref_to_self_map[self._ref]
        eventlog.log_d('done: fini video-work (ref:%s)' % self._ref)

        self._rm_session_fn(self._session)
        self._session = None
        raise VsError


class _VsService(IVsService):
    def __init__(self, twisted_reactor, args):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorUNIX
        :type args: VsServiceArgs
        """
        IVsService.__init__(self)
        self._reactor = twisted_reactor

        fs.silent_rm_r(args.tmp_path)
        fs.silent_mkdir_r(args.tmp_path)
        self._tmp_path = args.tmp_path

        self._image_pool = _VsImageSessionPool(args.store, args.chan, args.log_level, _IMAGE_POOL_MAX)
        self._video_pool = _VsVideoSessionPool(args.store, args.chan, args.log_level, _VIDEO_POOL_MAX)

    def destroy(self):
        del self._image_pool
        del self._video_pool

    def export_image_rgba(self, im_rgba, im_infos, bs_chan=None):
        """
        :type im_rgba: ImRgba
        :type im_infos: list[VsImageInfo]
        :type bs_chan: str | None

        :rtype: VsImageInfo | None
        :raise VsImageBufferError:
        :raise VsError:
        """
        session = self._image_pool.get_session()
        try:
            for im_info in im_infos:
                try:
                    image_w, image_h = session.export_image_rgba(im_info.ts / 1000, im_rgba.buffer, bs_chan=bs_chan)
                except VsSessionExpiredError:
                    raise

                except VsError:
                    eventlog.log_i('fail: export image-ts:%s (trying the next one)' % im_info.ts)
                    continue

                if image_w > im_rgba.w or image_h > im_rgba.h:
                    raise VsImageBufferError(image_w, image_h)

                elif image_w > 0 and image_h > 0:
                    im_rgba.reset(im_info.ts, image_w, image_h, im_rgba.dh)
                    return im_info

                else:
                    eventlog.log_e('fail: invalid image-wh:%sx%s' % (image_w, image_h))
                    raise VsError
        finally:
            self._image_pool.put_session(session)

    def export_image_jpeg(self, im_buffer, im_infos, bs_chan=None):
        """
        :type im_buffer: ImBuffer
        :type im_infos: list[VsImageInfo]
        :type bs_chan: str | None

        :rtype: VsImageInfo | None
        :raise VsImageBufferError2:
        :raise VsError:
        """
        session = self._image_pool.get_session()
        try:
            for im_info in im_infos:
                try:
                    image_size, image_ts = session.export_image_jpeg(im_info.ts / 1000, im_buffer.buffer, bs_chan=bs_chan)
                except VsSessionExpiredError:
                    raise

                except VsError:
                    eventlog.log_i('fail: export image-ts:%s (trying the next one)' % im_info.ts)
                    continue

                if image_size > im_buffer.buffer_size:
                    raise VsImageBufferError2(image_size)

                elif image_size > 0:
                    im_buffer.reset(image_ts * 1000, buffer_cap=image_size)
                    return im_info

                else:
                    eventlog.log_e('fail: invalid image-size:%s' % image_size)
                    raise VsError
        finally:
            self._image_pool.put_session(session)

    def export_video(self, min_ts, max_ts, max_w, max_h, rec_quality, blob_handler):
        """
        :type min_ts: long
        :type max_ts: long
        :type max_w: int
        :type max_h: int
        :type rec_quality: int
        :type blob_handler: (FileBlob | None) -> None

        :rtype: None
        :raise VsSessionAbsentError:
        :raise VsError:
        """
        video_work = None
        try:
            video_work = _VsVideoWork(
                self._reactor,
                self._tmp_path,
                lambda: self._video_pool.get_session(),
                lambda session: self._video_pool.put_session(session),
                min_ts, max_ts, max_w, max_h, rec_quality, blob_handler)

            video_work.run_async()

        except VsSessionAbsentError:
            eventlog.log_w('fail: session absent')
            if video_work:
                video_work.__del__()
            raise

        except VsError:
            eventlog.log_e('fail: session error')
            if video_work:
                video_work.__del__()
            raise
        except Exception:
            eventlog.log_x()
            if video_work:
                video_work.__del__()
            raise VsError


def init_service(twisted_reactor, args):
    """
    :type twisted_reactor: twisted.internet.interfaces.IReactorUNIX
    :type args: VsServiceArgs
    :rtype: IVsService
    """
    return _VsService(twisted_reactor, args)


def fini_service(service):
    """
    :type service: IVsService | _VsService | _StubService
    """
    service.destroy()


class _StubService(IVsService):
    def __init__(self):
        IVsService.__init__(self)

    def export_image_rgba(self, im_rgba, im_infos, bs_chan=None):
        """
        :type im_rgba: ImRgba
        :type im_infos: list[VsImageInfo]
        :type bs_chan: str | None

        :rtype: VsImageInfo | None
        :raise VsImageBufferError:
        :raise VsError:
        """
        return None

    def export_image_jpeg(self, im_buffer, im_infos, bs_chan=None):
        """
        :type im_buffer: ImBuffer
        :type im_infos: list[VsImageInfo]
        :type bs_chan: str | None

        :rtype: VsImageInfo | None
        :raise VsImageBufferError2:
        :raise VsError:
        """
        return None

    def export_video(self, min_ts, max_ts, max_w, max_h, rec_quality, blob_handler):
        """
        :type min_ts: long
        :type max_ts: long
        :type max_w: int
        :type max_h: int
        :type rec_quality: int
        :type blob_handler: (FileBlob | None) -> None

        :rtype: None
        :raise VsSessionAbsentError:
        :raise VsError:
        """
        return None

    def destroy(self):
        pass


def init_stub_service(twisted_reactor, args):
    """
    :type twisted_reactor: twisted.internet.interfaces.IReactorUNIX
    :type args: VsServiceArgs
    :rtype: IVsService
    """
    return _StubService()
